"""Meta tests for verifying ERROR vs FAILED status distinction."""

from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file, run_test_subprocess


@test()
async def test_assertion_failure_shows_as_failed() -> None:
    """Test that assertion failures (using assert_eq) show as FAILED, not ERROR."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import assert_eq

            @test()
            def test_will_fail_assertion() -> None:
                assert_eq(1, 2)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 1)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 1)


@test()
async def test_exception_shows_as_error() -> None:
    """Test that exceptions (like ValueError) show as ERROR, not FAILED."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test

            @test()
            def test_will_error() -> None:
                raise ValueError("Something went wrong")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 1)
    assert_eq(result["returncode"], 1)


@test()
async def test_bare_assert_shows_as_error() -> None:
    """Test that bare assert statements show as ERROR (not FAILED).

    This encourages users to use the assertion functions instead.
    """
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_eq

            @test()
            def test_with_bare_assert() -> None:
                assert_eq(1, 2, "Numbers don't match")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 1)
    assert_eq(result["returncode"], 1)


@test()
async def test_mixed_failures_and_errors() -> None:
    """Test that failed and error counts are tracked separately."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import assert_eq, assert_true

            @test()
            def test_assertion_failure_1() -> None:
                assert_eq(1, 2)

            @test()
            def test_assertion_failure_2() -> None:
                assert_true(False)

            @test()
            def test_error_1() -> None:
                raise ValueError("Error 1")

            @test()
            def test_error_2() -> None:
                raise TypeError("Error 2")

            @test()
            def test_error_3() -> None:
                1 / 0

            @test()
            def test_passing() -> None:
                assert_eq(1, 1)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 2)
    assert_eq(result["errors"], 3)
    assert_eq(result["returncode"], 1)


@test()
async def test_all_assertion_types_show_as_failed() -> None:
    """Test that all assertion helper functions result in FAILED status."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import (
                assert_eq,
                assert_ne,
                assert_true,
                assert_false,
                assert_in,
                assert_is,
                assert_is_none,
            )

            @test()
            def test_assert_eq_fail() -> None:
                assert_eq(1, 2)

            @test()
            def test_assert_ne_fail() -> None:
                assert_ne(1, 1)

            @test()
            def test_assert_true_fail() -> None:
                assert_true(False)

            @test()
            def test_assert_false_fail() -> None:
                assert_false(True)

            @test()
            def test_assert_in_fail() -> None:
                assert_in(5, [1, 2, 3])

            @test()
            def test_assert_is_fail() -> None:
                a = [1, 2]
                b = [1, 2]
                assert_is(a, b)

            @test()
            def test_assert_is_none_fail() -> None:
                assert_is_none(42)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 7)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 1)


@test()
async def test_various_exception_types_show_as_error() -> None:
    """Test that different exception types all show as ERROR."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test

            @test()
            def test_value_error() -> None:
                raise ValueError("Value error")

            @test()
            def test_type_error() -> None:
                raise TypeError("Type error")

            @test()
            def test_key_error() -> None:
                d = {}
                _ = d["missing_key"]

            @test()
            def test_index_error() -> None:
                lst = []
                _ = lst[0]

            @test()
            def test_zero_division() -> None:
                1 / 0
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 5)
    assert_eq(result["returncode"], 1)



from snektest.assertions import (
    assert_eq,
    assert_false,
    assert_ge,
    assert_gt,
    assert_in,
    assert_is,
    assert_is_none,
    assert_is_not,
    assert_is_not_none,
    assert_isinstance,
    assert_le,
    assert_len,
    assert_lt,
    assert_ne,
    assert_not_in,
    assert_not_isinstance,
    assert_raises,
    assert_true,
    fail,
)
from snektest.decorators import (
    Marker,
    load_fixture,
    session_fixture,
    test,
    test_hypothesis,
)
from snektest.models import Param as Param
from snektest.models import Scope as Scope
from snektest.models import UnreachableError as UnreachableError

__all__ = [
    "Marker",
    "Param",
    "Scope",
    "UnreachableError",
    "assert_eq",
    "assert_false",
    "assert_ge",
    "assert_gt",
    "assert_in",
    "assert_is",
    "assert_is_none",
    "assert_is_not",
    "assert_is_not_none",
    "assert_isinstance",
    "assert_le",
    "assert_len",
    "assert_lt",
    "assert_ne",
    "assert_not_in",
    "assert_not_isinstance",
    "assert_raises",
    "assert_true",
    "fail",
    "load_fixture",
    "session_fixture",
    "test",
    "test_hypothesis",
]



"""Shared fixtures for snektest's test suite."""

import tempfile
from collections.abc import Generator
from pathlib import Path

from snektest import session_fixture


@session_fixture()
def tmp_dir_fixture() -> Generator[Path]:
    """Fixture that provides a temporary directory for tests.

    Yields:
        Path to temporary directory
    """
    with tempfile.TemporaryDirectory() as tmp:
        yield Path(tmp)



from rich.console import Console

from snektest.models import (
    ErrorResult,
    FailedResult,
    PassedResult,
    TeardownFailure,
    TestResult,
)
from snektest.presenter.errors import print_failures as _print_failures
from snektest.presenter.summary import print_summary as _print_summary

console = Console()


def print_error(exc: str) -> None:
    """Print an error message in red."""
    console.print(exc, markup=False, style="red")


def print_test_result(result: TestResult) -> None:
    """Print the result of a single test."""
    console.print(
        f"{result.name!s} ... ", end="", markup=False, highlight=False, no_wrap=True
    )
    match result.result:
        case PassedResult():
            console.print(
                f"OK ({result.duration:.2f}s)",
                highlight=False,
                no_wrap=True,
                style="green",
                markup=False,
            )
        case FailedResult():
            console.print(
                f"FAIL ({result.duration:.2f}s)",
                highlight=False,
                no_wrap=True,
                style="red",
                markup=False,
            )
        case ErrorResult():
            console.print(
                f"ERROR ({result.duration:.2f}s)",
                highlight=False,
                no_wrap=True,
                style="dark_orange",
                markup=False,
            )


def print_failures(
    test_results: list[TestResult],
    session_teardown_failures: list[TeardownFailure] | None = None,
    session_teardown_output: str | None = None,
) -> None:
    """Print all failures."""
    _print_failures(
        console,
        test_results,
        session_teardown_failures=session_teardown_failures,
        session_teardown_output=session_teardown_output,
    )


def print_summary(
    test_results: list[TestResult],
    total_duration: float,
    session_teardown_failures: list[TeardownFailure] | None = None,
) -> None:
    """Print test summary."""
    _print_summary(
        console,
        test_results,
        total_duration,
        session_teardown_failures=session_teardown_failures,
    )


__all__ = [
    "console",
    "print_error",
    "print_failures",
    "print_summary",
    "print_test_result",
]



from collections.abc import Coroutine as _Coroutine
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Any, NewType

from pydantic import (
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    TypeAdapter,
    ValidationInfo,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema, PydanticCustomError
from pydantic_core.core_schema import (
    with_info_after_validator_function,
)

type Coroutine[T] = _Coroutine[None, None, T]


@dataclass
class PyFileType:
    def __get_pydantic_json_schema__(
        self, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        field_schema = handler(core_schema)
        field_schema.update(format="file-path", type="string")
        return field_schema

    def __get_pydantic_core_schema__(
        self, source: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return with_info_after_validator_function(
            self.validate_file,
            handler(source),
        )

    @staticmethod
    def validate_file(path: Path, _: ValidationInfo) -> Path:
        if not path.is_file():
            err_type = "path_not_file"
            msg = "Path does not point to a file"
            raise PydanticCustomError(err_type, msg)
        if path.suffix != ".py":
            err_type = "path_not_python"
            msg = "File path points to is not `.py`"
            raise PydanticCustomError(err_type, msg)
        return path

    def __hash__(self) -> int:
        return hash(self.__class__.__name__)


PyFilePath = Annotated[NewType("PyFile", Path), PyFileType()]
validate_PyFilePath = TypeAdapter[PyFilePath](PyFilePath).validate_python



import asyncio
import json
import sys
import threading
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import cast

from snektest.collection import TestsQueue, load_tests_from_filters
from snektest.decorators import Marker
from snektest.execution import run_tests
from snektest.models import (
    ArgsError,
    BadRequestError,
    CollectionError,
    ErrorResult,
    FailedResult,
    FilterItem,
    PassedResult,
    TeardownFailure,
    TestResult,
    UnreachableError,
)
from snektest.presenter import print_error


def _json_result_status(result: TestResult) -> str:
    match result.result:
        case PassedResult():
            return "passed"
        case FailedResult():
            return "failed"
        case ErrorResult():
            return "error"


def build_json_summary(summary: TestRunSummary) -> dict[str, object]:
    return {
        "passed": summary.passed,
        "failed": summary.failed,
        "errors": summary.errors,
        "fixture_teardown_failed": summary.fixture_teardown_failed,
        "session_teardown_failed": summary.session_teardown_failed,
        "tests": [
            {
                "name": str(result.name),
                "duration": result.duration,
                "markers": list(result.markers),
                "status": _json_result_status(result),
            }
            for result in summary.test_results
        ],
    }


@dataclass
class TestRunSummary:
    """Summary of test run results."""

    total_tests: int
    passed: int
    failed: int
    errors: int
    fixture_teardown_failed: int
    session_teardown_failed: int
    test_results: list[TestResult]
    session_teardown_failures: list[TeardownFailure]


@dataclass(frozen=True)
class CliOptions:
    capture_output: bool = True
    json_output: bool = False
    pdb_on_failure: bool = False
    mark: str | None = None


VALID_MARKER_VALUES = {marker.value for marker in Marker}


def _is_valid_mark_value(mark_value: str) -> bool:
    return mark_value in VALID_MARKER_VALUES


def _invalid_mark_message(mark_value: str) -> str:
    allowed = ", ".join(sorted(VALID_MARKER_VALUES))
    return f"Invalid --mark value: `{mark_value}`. Use one of: {allowed}"


def _parse_mark_value(
    argv: list[str], index: int, mark: str | None
) -> tuple[str | None, int] | int:
    if mark is not None:
        print_error("Only one --mark value is supported")
        return 2
    if index + 1 >= len(argv):
        print_error("Missing value for --mark")
        return 2
    mark_value = argv[index + 1]
    if mark_value.startswith("-"):
        print_error("Missing value for --mark")
        return 2
    if not _is_valid_mark_value(mark_value):
        print_error(_invalid_mark_message(mark_value))
        return 2
    return mark_value, index + 1


def parse_cli_args(argv: list[str]) -> tuple[list[str], CliOptions] | int:
    capture_output = True
    json_output = False
    pdb_on_failure = False
    mark: str | None = None
    potential_filter: list[str] = []

    index = 0
    while index < len(argv):
        command = argv[index]
        if command.startswith("-"):
            match command:
                case "-s":
                    capture_output = False
                case "--json-output":
                    json_output = True
                case "--pdb":
                    pdb_on_failure = True
                case "--mark":
                    parsed = _parse_mark_value(argv, index, mark)
                    if isinstance(parsed, int):
                        return parsed
                    mark, index = parsed
                case _:
                    print_error(f"Invalid option: `{command}`")
                    return 2
        else:
            potential_filter.append(command)
        index += 1

    if not potential_filter:
        potential_filter.append(".")

    options = CliOptions(
        capture_output=capture_output,
        json_output=json_output,
        pdb_on_failure=pdb_on_failure,
        mark=mark,
    )
    return potential_filter, options


async def _run_tests_with_producer_thread(
    filter_items: list[FilterItem],
    *,
    capture_output: bool,
    pdb_on_failure: bool,
    mark: str | None = None,
) -> tuple[list[TestResult], list[TeardownFailure]]:
    queue = TestsQueue()
    collection_exception: list[BaseException] = []

    producer_thread = threading.Thread(
        target=load_tests_from_filters,
        kwargs={
            "filter_items": filter_items,
            "queue": queue,
            "loop": asyncio.get_running_loop(),
            "mark": mark,
            "exception_holder": collection_exception,
        },
    )

    producer_thread.start()

    try:
        test_results, session_teardown_failures = await run_tests(
            queue=queue,
            capture_output=capture_output,
            pdb_on_failure=pdb_on_failure,
        )
    finally:
        producer_thread.join()
        if collection_exception:
            raise collection_exception[0]

    return test_results, session_teardown_failures


def exit_code_from_summary(summary: TestRunSummary) -> int:
    has_failures = (
        summary.failed > 0
        or summary.errors > 0
        or summary.fixture_teardown_failed > 0
        or summary.session_teardown_failed > 0
    )
    return 1 if has_failures else 0


async def run_tests_programmatic(
    filter_items: list[FilterItem],
    *,
    capture_output: bool = True,
    pdb_on_failure: bool = False,
    mark: str | None = None,
) -> TestRunSummary:
    """Run tests and return structured results instead of printing.

    This is the programmatic API for testing snektest itself.
    Returns structured data instead of just printing and exiting.

    Args:
        filter_items: List of filter items to run tests from
        capture_output: Whether to capture test output

    Returns:
        TestRunSummary with test results and counts
    """
    if mark is not None and not _is_valid_mark_value(mark):
        raise BadRequestError(_invalid_mark_message(mark))

    test_results, session_teardown_failures = await _run_tests_with_producer_thread(
        filter_items,
        capture_output=capture_output,
        pdb_on_failure=pdb_on_failure,
        mark=mark,
    )

    return TestRunSummary(
        total_tests=len(test_results),
        passed=sum(1 for r in test_results if isinstance(r.result, PassedResult)),
        failed=sum(1 for r in test_results if isinstance(r.result, FailedResult)),
        errors=sum(1 for r in test_results if isinstance(r.result, ErrorResult)),
        fixture_teardown_failed=sum(
            1 for r in test_results if r.fixture_teardown_failures
        ),
        session_teardown_failed=len(session_teardown_failures),
        test_results=test_results,
        session_teardown_failures=session_teardown_failures,
    )


async def run_script(
    argv: list[str] | None = None,
    *,
    run_tests_programmatic_fn: Callable[..., Coroutine[object, object, object]]
    | None = None,
) -> int:
    """Parse arguments and run tests."""
    parsed = parse_cli_args(sys.argv[1:] if argv is None else argv)
    if isinstance(parsed, int):
        return parsed

    potential_filter, options = parsed

    try:
        filter_items = [FilterItem(item) for item in potential_filter]
    except ArgsError as e:
        print_error(str(e))
        return 2

    runner = run_tests_programmatic_fn or run_tests_programmatic
    try:
        summary = cast(
            "TestRunSummary",
            await runner(
                filter_items,
                capture_output=options.capture_output,
                pdb_on_failure=options.pdb_on_failure,
                mark=options.mark,
            ),
        )
    except asyncio.CancelledError:
        return 2

    if options.json_output:
        print(json.dumps(build_json_summary(summary)))

    return exit_code_from_summary(summary)


def main() -> None:
    """Main entry point for the CLI."""
    async_runner = cast("Callable[[object], int]", asyncio.run)
    sys.exit(main_inner(async_runner=async_runner))


def main_inner(
    *,
    async_runner: Callable[[object], int],
    argv: list[str] | None = None,
) -> int:
    try:
        coroutine = run_script(argv)
        return async_runner(coroutine)
    except CollectionError as e:
        print_error(f"Collection error: {e}")
        return 2
    except BadRequestError as e:
        print_error(f"Bad request error: {e}")
        return 2
    except UnreachableError as e:
        print_error(f"Internal error: {e}")
        return 2
    except KeyboardInterrupt:
        print_error("Interrupted by user")
        return 2
    except Exception as e:
        print_error(f"Unexpected error: {e}")
        return 1


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from enum import Enum, auto
from io import StringIO
from itertools import product
from pathlib import Path
from types import TracebackType
from typing import Any


class CollectionError(BaseException): ...


class ArgsError(BaseException): ...


class UnreachableError(BaseException): ...


class BadRequestError(BaseException):
    """When user didn't write test code correctly"""


class AssertionFailure(AssertionError):  # noqa: N818
    def __init__(
        self,
        message: str,
        *,
        actual: Any = None,
        expected: Any = None,
        operator: str | None = None,
    ) -> None:
        super().__init__(message)
        self.actual = actual
        self.expected = expected
        self.operator = operator


SnektestError = CollectionError | ArgsError | UnreachableError | AssertionFailure


class FilterItem:
    """Represents a single test filter item"""

    def __init__(self, raw_input: str) -> None:
        if "::" not in raw_input:
            path = Path(raw_input)
            function_name = None
            params = None
        else:
            file_part, rest = raw_input.split("::", 1)
            if rest == "":
                msg = f"Invalid test filter - nothing given after semicolon in '{raw_input}'"
                raise ArgsError(msg)

            path = Path(file_part)

            if "[" in rest:
                if not rest.endswith("]"):
                    msg = f"Invalid test filter - unterminated `[` in '{raw_input}'"
                    raise ArgsError(msg)
                rest = rest.removesuffix("]")
                function_name, params = rest.split("[", 1)
            else:
                function_name = rest
                params = None

        if not path.exists():
            msg = f"Invalid test filter - provided path does not exist in '{raw_input}'"
            raise ArgsError(msg)

        if path.is_file() and path.suffix != ".py":
            msg = f"Invalid test filter - file is not a Python script in '{raw_input}'"
            raise ArgsError(msg)

        if path.is_file() and not path.name.startswith("test_"):
            msg = (
                f"Invalid test filter - file does not start with _test in '{raw_input}'"
            )
            raise ArgsError(msg)

        if function_name is not None and not function_name.isidentifier():
            msg = f"Invalid test filter - invalid identifier {function_name} in '{raw_input}'"
            raise ArgsError(msg)

        self.file_path = path
        self.function_name = function_name
        self.params = params

    def __str__(self) -> str:
        result = str(self.file_path)
        if self.function_name is not None:
            result += f"::{self.function_name}"
        if self.params:
            result += f"[{self.params}]"
        return result

    def __repr__(self) -> str:
        return f"FilterItem(file_path={self.file_path!r}, function_name={self.function_name!r}, params={self.params!r})"


# Set kw_only so we can write attributes in the order they appear
@dataclass(kw_only=True)
class TestName:
    file_path: Path
    func_name: str
    params_part: str

    def __str__(self) -> str:
        result = str(self.file_path)
        result += f"::{self.func_name}"
        if self.params_part:
            result += f"[{self.params_part}]"
        return result


class PassedResult: ...


@dataclass
class Param[T]:
    value: T
    name: str

    @staticmethod
    def to_dict(
        params: tuple[list[Param[Any]], ...],
    ) -> dict[str, tuple[Param[Any], ...]]:
        """Create a dictionary that contains all possible params combinations.

        For tests with no parameters, returns {"": ()} to ensure the test runs once.
        """
        if not params:
            return {"": ()}

        combinations = product(*params)
        result: dict[str, tuple[Param[Any], ...]] = {}
        for combination in combinations:
            result[", ".join([param.name for param in combination])] = combination
        return result


class Scope(Enum):
    FUNCTION = auto()
    SESSION = auto()


@dataclass(frozen=True)
class FailedResult:
    exc_type: type[BaseException]
    exc_value: BaseException
    traceback: TracebackType


@dataclass(frozen=True)
class ErrorResult:
    exc_type: type[BaseException]
    exc_value: BaseException
    traceback: TracebackType


@dataclass(frozen=True)
class TeardownFailure:
    """Represents a fixture teardown failure"""

    fixture_name: str
    exc_type: type[BaseException]
    exc_value: BaseException
    traceback: TracebackType


@dataclass
class TestResult:
    name: TestName
    duration: float
    result: PassedResult | FailedResult | ErrorResult
    markers: tuple[str, ...]
    captured_output: StringIO
    fixture_teardown_failures: list[TeardownFailure]
    fixture_teardown_output: str | None
    warnings: list[str]



import asyncio
from collections.abc import Callable
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getmembers, isfunction
from pathlib import Path
from sys import modules
from types import FunctionType
from typing import TypeGuard, cast

from pydantic import ValidationError

from snektest.annotations import PyFilePath, validate_PyFilePath
from snektest.models import CollectionError, FilterItem, TestName
from snektest.utils import (
    get_test_function_markers,
    get_test_function_params,
    is_test_function,
)

TEST_FILE_PREFIX = "test_"

FuncTestEntry = tuple[TestName, FunctionType]
TestsQueue = asyncio.Queue[FuncTestEntry]


def load_tests_from_file(  # noqa: PLR0913
    file_path: PyFilePath,
    filter_item: FilterItem,
    queue: TestsQueue,
    loop: asyncio.AbstractEventLoop,
    *,
    mark: str | None = None,
    spec_loader: Callable[..., object] = spec_from_file_location,
) -> None:
    """Load and queue tests from a single Python file."""
    module_name = ".".join(file_path.with_suffix("").parts)
    if module_name in modules:
        module = modules[module_name]
    else:
        spec = spec_loader(module_name, file_path)
        spec_value = cast("ModuleSpec", spec)
        loader = getattr(spec_value, "loader", None)
        if loader is None:
            msg = f"Could not load spec from {file_path}"
            raise CollectionError(msg)

        module = module_from_spec(spec_value)
        modules[module_name] = module
        loader.exec_module(module)

    runnable_functions = [func for _, func in getmembers(module, isfunction)]
    runnable_functions = filter(is_test_function, runnable_functions)
    if filter_item.function_name:
        runnable_functions = filter(
            lambda func: func.__name__ == filter_item.function_name, runnable_functions
        )

    if mark is not None:
        runnable_functions = filter(
            lambda func: mark in get_test_function_markers(func), runnable_functions
        )

    for func in runnable_functions:
        for param_names in get_test_function_params(func):
            if filter_item.params and filter_item.params != param_names:
                continue
            test_name = TestName(
                file_path=file_path, func_name=func.__name__, params_part=param_names
            )
            _ = loop.call_soon_threadsafe(queue.put_nowait, (test_name, func))


def generate_file_list(filter_item: FilterItem) -> list[PyFilePath]:
    """Generate a list of valid file paths for given filter item."""

    def path_is_runnable(file_path: Path) -> TypeGuard[PyFilePath]:
        if not file_path.name.startswith(TEST_FILE_PREFIX):
            return False
        try:
            file_path = validate_PyFilePath(file_path)
        except ValidationError:
            return False
        return True

    if filter_item.file_path.is_dir():
        paths = [
            dirpath / name
            for dirpath, _, filenames in filter_item.file_path.walk()
            for name in filenames
        ]
    else:
        paths = [filter_item.file_path]

    return [path for path in paths if path_is_runnable(path)]


def load_tests_from_filters(
    filter_items: list[FilterItem],
    queue: TestsQueue,
    loop: asyncio.AbstractEventLoop,
    *,
    mark: str | None = None,
    exception_holder: list[BaseException] | None = None,
) -> None:
    """Load tests from all filter items and populate the queue.

    Args:
        filter_items: List of filter items to load tests from
        queue: Queue to populate with tests
        loop: Event loop for thread-safe queue operations
        exception_holder: Optional list to store exception if one occurs during collection
    """
    try:
        for filter_item in filter_items:
            file_paths = generate_file_list(filter_item)
            for file_path in file_paths:
                load_tests_from_file(
                    file_path=file_path,
                    filter_item=filter_item,
                    queue=queue,
                    loop=loop,
                    mark=mark,
                )
    except BaseException as e:
        if exception_holder is not None:
            exception_holder.append(e)
    finally:
        _ = loop.call_soon_threadsafe(queue.shutdown)



"""Meta tests for fixture error handling."""

from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file, run_test_subprocess


@test()
def test_function_fixture_teardown_failure() -> None:
    """Test that function fixture teardown failures are reported."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import Generator
            from snektest import load_fixture, test

            def fixture_with_failing_teardown() -> Generator[None]:
                yield None
                msg = "failing teardown"
                raise ValueError(msg)

            @test()
            def test_with_bad_fixture() -> None:
                _ = load_fixture(fixture_with_failing_teardown())
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["fixture_teardown_failed"], 1)


@test()
def test_session_fixture_teardown_failure() -> None:
    """Test that session fixture teardown failures are reported."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import Generator
            from snektest import load_fixture, session_fixture, test

            @session_fixture()
            def session_fixture_with_failing_teardown() -> Generator[None]:
                yield None
                msg = "failing teardown"
                raise ValueError(msg)

            @test()
            def test_with_bad_session_fixture() -> None:
                _ = load_fixture(session_fixture_with_failing_teardown())
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["session_teardown_failed"], 1)


@test()
def test_good_fixture_no_errors() -> None:
    """Test that good fixtures don't report any errors."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import Generator
            from snektest import load_fixture, test

            def good_fixture() -> Generator[None]:
                yield None

            @test()
            def test_with_good_fixture() -> None:
                _ = load_fixture(good_fixture())
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["fixture_teardown_failed"], 0)
    assert_eq(result["session_teardown_failed"], 0)



from __future__ import annotations

import asyncio
import contextlib
import json
import runpy
import sys
from collections.abc import Callable
from io import StringIO
from pathlib import Path
from typing import cast

from snektest import assert_eq, assert_raises, test
from snektest.cli import (
    main,
    main_inner,
    parse_cli_args,
    run_script,
    run_tests_programmatic,
)
from snektest.models import (
    BadRequestError,
    CollectionError,
    FilterItem,
    PassedResult,
    TestName,
    TestResult,
    UnreachableError,
)


@test()
def test_parse_cli_args_invalid_option_returns_2() -> None:
    result = parse_cli_args(["--nope"])
    assert_eq(result, 2)


@test()
def test_parse_cli_args_defaults_to_dot() -> None:
    parsed = parse_cli_args([])
    assert not isinstance(parsed, int)

    potential_filter, options = parsed
    assert_eq(potential_filter, ["."])
    assert_eq(options.capture_output, True)
    assert_eq(options.json_output, False)
    assert_eq(options.pdb_on_failure, False)
    assert_eq(options.mark, None)


@test()
def test_parse_cli_args_s_flag_disables_capture() -> None:
    parsed = parse_cli_args(["-s", "."])
    assert not isinstance(parsed, int)

    _potential_filter, options = parsed
    assert_eq(options.capture_output, False)
    assert_eq(options.mark, None)


@test()
async def test_run_script_returns_2_on_args_error() -> None:
    result = await run_script(["does-not-exist"])
    assert_eq(result, 2)


@test()
async def test_run_script_returns_parse_cli_args_exit_code() -> None:
    result = await run_script(["--nope"])
    assert_eq(result, 2)


@test()
async def test_run_script_returns_2_on_cancelled_error() -> None:
    async def raise_cancelled(*args: object, **kwargs: object) -> object:
        _ = (args, kwargs)
        raise asyncio.CancelledError

    result = await run_script(["."], run_tests_programmatic_fn=raise_cancelled)
    assert_eq(result, 2)


@test()
async def test_run_tests_programmatic_rejects_unknown_marker() -> None:
    with assert_raises(BadRequestError):
        _ = await run_tests_programmatic([FilterItem(".")], mark="needs-s3")


@test()
async def test_run_script_json_output_includes_markers() -> None:
    async def fake_run(*args: object, **kwargs: object) -> object:
        _ = (args, kwargs)
        test_result = TestResult(
            name=TestName(
                file_path=Path("tests/test_fake.py"), func_name="t", params_part=""
            ),
            duration=0.0,
            result=PassedResult(),
            markers=("fast",),
            captured_output=StringIO(""),
            fixture_teardown_failures=[],
            fixture_teardown_output=None,
            warnings=[],
        )
        return type(
            "Summary",
            (),
            {
                "passed": 1,
                "failed": 0,
                "errors": 0,
                "fixture_teardown_failed": 0,
                "session_teardown_failed": 0,
                "test_results": [test_result],
            },
        )()

    buffer = StringIO()
    with contextlib.redirect_stdout(buffer):
        result = await run_script(
            ["--json-output"],
            run_tests_programmatic_fn=fake_run,
        )
    assert_eq(result, 0)
    payload = json.loads(buffer.getvalue())
    assert_eq(payload["tests"][0]["markers"], ["fast"])


@test()
def test_main_runs_and_exits() -> None:
    original_argv = list(sys.argv)
    try:
        sys.argv = ["snektest", "--nope"]
        with assert_raises(SystemExit):
            main()
    finally:
        sys.argv = original_argv


@test()
def test_run_module_main_invokes_cli() -> None:
    original_argv = list(sys.argv)
    try:
        sys.argv = ["snektest", "--nope"]
        with assert_raises(SystemExit):
            _ = runpy.run_module("snektest.cli", run_name="__main__")
    finally:
        sys.argv = original_argv


@test()
def test_main_exit_paths() -> None:
    def assert_exit_code(exc: BaseException | None, expected: int) -> None:
        def fake_run(coro: object) -> int:
            closer_obj = getattr(coro, "close", None)
            if callable(closer_obj):
                closer = cast("Callable[[], None]", closer_obj)
                closer()
            if exc is not None:
                raise exc
            return 0

        result = main_inner(async_runner=fake_run, argv=["."])
        assert_eq(result, expected)
        result = main_inner(async_runner=fake_run, argv=["."])
        assert_eq(result, expected)
        result = main_inner(async_runner=fake_run, argv=["."])
        assert_eq(result, expected)

    assert_exit_code(None, 0)
    assert_exit_code(CollectionError("x"), 2)
    assert_exit_code(BadRequestError("x"), 2)
    assert_exit_code(UnreachableError("x"), 2)
    assert_exit_code(KeyboardInterrupt(), 2)
    assert_exit_code(RuntimeError("x"), 1)



"""Meta tests for handling import errors in test collection."""

import subprocess
import sys
from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_ne, fail
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file


@test()
def test_import_error_does_not_hang() -> None:
    """Test that import errors don't cause the test runner to hang.

    When a test file fails to import (e.g., raises an exception at module level),
    the producer thread should handle the error gracefully and shut down the queue,
    rather than hanging forever.
    """
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_true

            raise RuntimeError("Intentional import error for testing")

            @test()
            def test_unreachable() -> None:
                assert_true(True)
        """),
    )

    try:
        cmd = [sys.executable, "-m", "snektest.cli", "--json-output", str(test_file)]
        result = subprocess.run(
            cmd,
            check=False,
            capture_output=True,
            text=True,
            timeout=0.5,
        )
        assert_ne(result.returncode, 0)
    except subprocess.TimeoutExpired:
        fail("Test runner hung on import error")



"""Meta tests for running Hypothesis-based tests via snektest."""

from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from snektest.decorators import Marker
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file, run_test_subprocess


@test()
async def test_hypothesis_sync_test_passes() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent(
            """
            from hypothesis import Phase, settings
            from hypothesis import strategies as st

            from snektest import assert_eq, test_hypothesis


            @settings(max_examples=1, phases=[Phase.generate], database=None, deadline=None)
            @test_hypothesis(st.just(0))
            def test_prop(x: int) -> None:
                assert_eq(x, 0)
            """
        ),
        name="test_hypothesis_sync_pass",
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 0)


@test()
async def test_hypothesis_async_test_passes() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent(
            """
            from hypothesis import Phase, settings
            from hypothesis import strategies as st

            from snektest import assert_eq, test_hypothesis


            @settings(max_examples=1, phases=[Phase.generate], database=None, deadline=None)
            @test_hypothesis(st.just(0))
            async def test_prop(x: int) -> None:
                assert_eq(x, 0)
            """
        ),
        name="test_hypothesis_async_pass",
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 0)


@test()
async def test_hypothesis_failure_counts_as_failed() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent(
            """
            from hypothesis import Phase, settings
            from hypothesis import strategies as st

            from snektest import assert_gt, test_hypothesis


            @settings(max_examples=1, phases=[Phase.generate], database=None, deadline=None)
            @test_hypothesis(st.just(0))
            def test_prop(x: int) -> None:
                assert_gt(x, 0)
            """
        ),
        name="test_hypothesis_fail",
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 0)
    assert_eq(result["failed"], 1)
    assert_eq(result["errors"], 0)
    assert_eq(result["returncode"], 1)


@test()
async def test_hypothesis_marked_test_filters_with_cli_mark() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent(
            """
            from hypothesis import Phase, settings
            from hypothesis import strategies as st

            from snektest import Marker, assert_eq, test_hypothesis


            @settings(max_examples=1, phases=[Phase.generate], database=None, deadline=None)
            @test_hypothesis(st.just(0), mark=Marker.FAST)
            async def test_fast(x: int) -> None:
                assert_eq(x, 0)


            @settings(max_examples=1, phases=[Phase.generate], database=None, deadline=None)
            @test_hypothesis(st.just(0), mark=Marker.SLOW)
            def test_slow(x: int) -> None:
                assert_eq(x, 0)
            """
        ),
        name="test_hypothesis_mark_filter",
    )

    payload = run_test_subprocess(test_file, "--mark", Marker.FAST.value)
    assert_eq(payload["passed"], 1)
    assert_eq(payload["failed"], 0)
    assert_eq(payload["errors"], 0)
    assert_eq(payload["tests"][0]["name"], f"{test_file}::test_fast")
    assert_eq(payload["tests"][0]["markers"], [Marker.FAST.value])
    assert_eq(payload["returncode"], 0)



"""Unit tests for snektest models."""

from pathlib import Path
from tempfile import TemporaryDirectory

from snektest import test
from snektest.assertions import assert_eq, assert_raises
from snektest.models import ArgsError, FilterItem


@test()
def test_filter_item_directory() -> None:
    """Test FilterItem with a directory path."""
    filter_item = FilterItem("tests")
    assert_eq(filter_item.file_path, Path("tests"))
    assert_eq(filter_item.function_name, None)
    assert_eq(filter_item.params, None)


@test()
def test_filter_item_file() -> None:
    """Test FilterItem with a file path."""
    filter_item = FilterItem("tests/isolated/test_basic.py")
    assert_eq(filter_item.file_path, Path("tests/isolated/test_basic.py"))
    assert_eq(filter_item.function_name, None)
    assert_eq(filter_item.params, None)


@test()
def test_filter_item_with_function() -> None:
    """Test FilterItem with file and function name."""
    filter_item = FilterItem("tests/isolated/test_basic.py::test_assert_eq_passes")
    assert_eq(filter_item.file_path, Path("tests/isolated/test_basic.py"))
    assert_eq(filter_item.function_name, "test_assert_eq_passes")
    assert_eq(filter_item.params, None)


@test()
def test_filter_item_with_params() -> None:
    """Test FilterItem with file, function, and params."""
    filter_item = FilterItem("tests/isolated/test_basic.py::test_func[param1, param2]")
    assert_eq(filter_item.file_path, Path("tests/isolated/test_basic.py"))
    assert_eq(filter_item.function_name, "test_func")
    assert_eq(filter_item.params, "param1, param2")


@test()
def test_filter_item_str_simple() -> None:
    """Test FilterItem string representation for simple path."""
    filter_item = FilterItem("tests")
    assert_eq(str(filter_item), "tests")


@test()
def test_filter_item_str_with_function() -> None:
    """Test FilterItem string representation with function."""
    filter_item = FilterItem("tests/isolated/test_basic.py::test_func")
    assert_eq(str(filter_item), "tests/isolated/test_basic.py::test_func")


@test()
def test_filter_item_str_with_params() -> None:
    """Test FilterItem string representation with params."""
    filter_item = FilterItem("tests/isolated/test_basic.py::test_func[params]")
    assert_eq(str(filter_item), "tests/isolated/test_basic.py::test_func[params]")


@test()
def test_filter_item_nonexistent_path() -> None:
    """Test FilterItem raises error for nonexistent path."""
    with assert_raises(ArgsError) as exc_info:
        _ = FilterItem("nonexistent/path/to/file.py")

    msg = str(exc_info.exception)
    assert_eq("provided path does not exist" in msg, True)


@test()
def test_filter_item_non_python_file() -> None:
    """Test FilterItem raises error for non-Python file."""
    with TemporaryDirectory() as tmpdir:
        test_file = Path(tmpdir) / "test_file.txt"
        _ = test_file.write_text("content")

        with assert_raises(ArgsError) as exc_info:
            _ = FilterItem(str(test_file))

        msg = str(exc_info.exception)
        assert_eq("file is not a Python script" in msg, True)


@test()
def test_filter_item_file_not_starting_with_test() -> None:
    """Test FilterItem raises error for Python file not starting with test_."""
    with TemporaryDirectory() as tmpdir:
        test_file = Path(tmpdir) / "myfile.py"
        _ = test_file.write_text("# python code")

        with assert_raises(ArgsError) as exc_info:
            _ = FilterItem(str(test_file))

        msg = str(exc_info.exception)
        assert_eq("file does not start with _test" in msg, True)


@test()
def test_filter_item_empty_after_double_colon() -> None:
    """Test FilterItem raises error when nothing follows ::."""
    with assert_raises(ArgsError) as exc_info:
        _ = FilterItem("tests/isolated/test_basic.py::")

    msg = str(exc_info.exception)
    assert_eq("nothing given after semicolon" in msg, True)


@test()
def test_filter_item_unterminated_bracket() -> None:
    """Test FilterItem raises error for unterminated bracket."""
    with assert_raises(ArgsError) as exc_info:
        _ = FilterItem("tests/isolated/test_basic.py::test_func[param")

    msg = str(exc_info.exception)
    assert_eq("unterminated" in msg, True)


@test()
def test_filter_item_invalid_identifier() -> None:
    """Test FilterItem raises error for invalid function name."""
    with assert_raises(ArgsError) as exc_info:
        _ = FilterItem("tests/isolated/test_basic.py::123invalid")

    msg = str(exc_info.exception)
    assert_eq("invalid identifier" in msg, True)


@test()
def test_filter_item_repr() -> None:
    """Test FilterItem repr includes all components."""
    filter_item = FilterItem("tests/isolated/test_basic.py::test_func[params]")
    repr_str = repr(filter_item)

    assert_eq("FilterItem" in repr_str, True)
    assert_eq("test_basic.py" in repr_str, True)
    assert_eq("test_func" in repr_str, True)
    assert_eq("params" in repr_str, True)



from __future__ import annotations

import inspect
import pdb  # noqa: T100
import sys
import warnings
from collections.abc import Callable, Generator
from contextlib import contextmanager
from dataclasses import dataclass
from io import StringIO
from typing import Any, TextIO


class StdinProxy:
    """Proxy for sys.stdin that disables output capture when read from.

    This allows breakpoint()/pdb to work correctly even when output is being
    captured. When stdin is read from (e.g., pdb waiting for user input),
    the proxy calls a callback to restore stdout/stderr so the debugger
    can display its prompt and receive user input.
    """

    def __init__(
        self,
        original_stdin: TextIO,
        disable_capture_callback: Callable[[], None],
    ) -> None:
        self._original_stdin = original_stdin
        self._disable_capture = disable_capture_callback

    def read(self, size: int = -1) -> str:
        self._disable_capture()
        return self._original_stdin.read(size)

    def readline(self, size: int = -1) -> str:
        self._disable_capture()
        return self._original_stdin.readline(size)

    def readlines(self, hint: int = -1) -> list[str]:
        self._disable_capture()
        return self._original_stdin.readlines(hint)

    def __iter__(self) -> StdinProxy:
        self._disable_capture()
        return self

    def __next__(self) -> str:
        self._disable_capture()
        return next(self._original_stdin)

    def fileno(self) -> int:
        return self._original_stdin.fileno()

    def isatty(self) -> bool:
        return self._original_stdin.isatty()

    def close(self) -> None:
        pass  # Don't close the original stdin

    @property
    def closed(self) -> bool:
        return self._original_stdin.closed

    @property
    def encoding(self) -> str:
        return self._original_stdin.encoding

    @property
    def mode(self) -> str:
        return self._original_stdin.mode

    @property
    def name(self) -> str:
        return self._original_stdin.name

    def __getattr__(self, name: str) -> object:
        return getattr(self._original_stdin, name)


@dataclass(frozen=True)
class _OriginalSysState:
    stdout: Any
    stderr: Any
    stdin: Any
    settrace: Any
    breakpointhook: Any


def _make_disable_capture(
    *,
    system_stdout: Any,
    system_stderr: Any,
) -> Callable[[], None]:
    capture_disabled = False

    def disable_capture() -> None:
        nonlocal capture_disabled
        if capture_disabled:
            return
        sys.stdout = system_stdout
        sys.stderr = system_stderr
        capture_disabled = True

    return disable_capture


def _make_settrace_wrapper(
    *,
    system_settrace: Any,
    disable_capture: Callable[[], None],
) -> Callable[[Any], None]:
    def settrace_wrapper(func: Any) -> None:
        disable_capture()
        system_settrace(func)

    return settrace_wrapper


def _maybe_run_inline_pdb_breakpoint(
    *,
    system_breakpointhook: Any,
    caller_frame: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    pdb_factory: Callable[..., Any] = pdb.Pdb,
) -> Any:
    if system_breakpointhook is not sys.__breakpointhook__:
        return system_breakpointhook(*args, **kwargs)

    if caller_frame is None:
        return system_breakpointhook(*args, **kwargs)
    if args:
        return system_breakpointhook(*args, **kwargs)

    header = kwargs.get("header")
    commands = kwargs.get("commands")
    if set(kwargs) - {"header", "commands"}:
        return system_breakpointhook(*args, **kwargs)

    debugger: pdb.Pdb = pdb_factory(
        mode="inline",
        backend="monitoring",
        colorize=True,
    )
    if header is not None:
        _ = debugger.message(header)
    return debugger.set_trace(caller_frame, commands=commands)


def _make_breakpointhook_wrapper(
    *,
    system_breakpointhook: Any,
    disable_capture: Callable[[], None],
    frame_provider: Callable[[], Any] = inspect.currentframe,
    pdb_factory: Callable[..., Any] = pdb.Pdb,
) -> Callable[..., Any]:
    def breakpointhook_wrapper(*args: Any, **kwargs: Any) -> Any:
        disable_capture()
        sys.breakpointhook = system_breakpointhook

        frame = frame_provider()
        caller_frame = frame.f_back if frame else None

        return _maybe_run_inline_pdb_breakpoint(
            system_breakpointhook=system_breakpointhook,
            caller_frame=caller_frame,
            args=args,
            kwargs=kwargs,
            pdb_factory=pdb_factory,
        )

    return breakpointhook_wrapper


def _format_warnings(warnings_list: list[warnings.WarningMessage]) -> list[str]:
    return [
        f"{warning.filename}:{warning.lineno}: {warning.category.__name__}: {warning.message}"
        for warning in warnings_list
    ]


def _install_capture(
    *,
    output_buffer: StringIO,
    system_stdin: Any,
    disable_capture: Callable[[], None],
    settrace_wrapper: Any,
    breakpointhook_wrapper: Any,
) -> None:
    sys.stdout = output_buffer
    sys.stderr = output_buffer
    sys.stdin = StdinProxy(system_stdin, disable_capture)
    sys.settrace = settrace_wrapper
    sys.breakpointhook = breakpointhook_wrapper


def _restore_system_state(system: _OriginalSysState) -> None:
    sys.stdout = system.stdout
    sys.stderr = system.stderr
    sys.stdin = system.stdin
    sys.settrace = system.settrace
    sys.breakpointhook = system.breakpointhook


@contextmanager
def capture_output(
    *,
    frame_provider: Callable[[], Any] = inspect.currentframe,
    pdb_factory: Callable[..., Any] = pdb.Pdb,
) -> Generator[tuple[StringIO, list[str]]]:
    """Context manager to capture stdout, stderr, and warnings.

    If stdin is read from (e.g., by pdb/breakpoint), output capture is
    automatically disabled so the debugger can function properly.
    """
    output_buffer = StringIO()
    captured_warnings: list[str] = []

    original_sys = _OriginalSysState(
        stdout=sys.stdout,
        stderr=sys.stderr,
        stdin=sys.stdin,
        settrace=sys.settrace,
        breakpointhook=sys.breakpointhook,
    )

    disable_capture = _make_disable_capture(
        system_stdout=original_sys.stdout,
        system_stderr=original_sys.stderr,
    )
    settrace_wrapper = _make_settrace_wrapper(
        system_settrace=original_sys.settrace,
        disable_capture=disable_capture,
    )
    breakpointhook_wrapper = _make_breakpointhook_wrapper(
        system_breakpointhook=original_sys.breakpointhook,
        disable_capture=disable_capture,
        frame_provider=frame_provider,
        pdb_factory=pdb_factory,
    )

    _install_capture(
        output_buffer=output_buffer,
        system_stdin=original_sys.stdin,
        disable_capture=disable_capture,
        settrace_wrapper=settrace_wrapper,
        breakpointhook_wrapper=breakpointhook_wrapper,
    )

    with warnings.catch_warnings(record=True) as warning_list:
        warnings.simplefilter("always")
        try:
            yield output_buffer, captured_warnings
        finally:
            captured_warnings.extend(_format_warnings(warning_list))
            _restore_system_state(original_sys)


@contextmanager
def maybe_capture_output(
    capture: bool,
) -> Generator[tuple[StringIO, list[str]]]:
    """Conditionally capture output based on a flag."""
    if capture:
        with capture_output() as (buffer, warnings_list):
            yield buffer, warnings_list
    else:
        buffer = StringIO()
        warnings_list: list[str] = []
        yield buffer, warnings_list



from __future__ import annotations

from dataclasses import dataclass
from typing import cast

from rich.console import Console

from snektest.models import (
    AssertionFailure,
    ErrorResult,
    FailedResult,
    TeardownFailure,
    TestResult,
)
from snektest.presenter.diff import render_assertion_failure
from snektest.presenter.traceback import render_traceback


@dataclass(frozen=True)
class FailureGroups:
    failures: list[TestResult]
    errors: list[TestResult]
    fixture_teardown_failures: list[TestResult]


def _collect_failure_groups(test_results: list[TestResult]) -> FailureGroups:
    failures = [
        result for result in test_results if isinstance(result.result, FailedResult)
    ]
    errors = [
        result for result in test_results if isinstance(result.result, ErrorResult)
    ]
    fixture_teardown_failures = [
        result for result in test_results if result.fixture_teardown_failures
    ]
    return FailureGroups(
        failures=failures,
        errors=errors,
        fixture_teardown_failures=fixture_teardown_failures,
    )


def _print_optional_output(console: Console, *, title: str, output: str | None) -> None:
    if not output:
        return
    console.print()
    console.print(title, style="yellow", markup=False)
    console.print(output, markup=False, highlight=False)


def _print_result_details(
    console: Console,
    *,
    result: TestResult,
    outcome: FailedResult | ErrorResult,
) -> None:
    exc_type = outcome.exc_type
    exc_value = outcome.exc_value
    traceback = outcome.traceback

    if isinstance(exc_value, AssertionFailure):
        render_traceback(
            console,
            exc_type,
            exc_value,
            traceback,
            show_exception_line=False,
        )
        render_assertion_failure(console, exc_value)
    else:
        render_traceback(console, exc_type, exc_value, traceback)

    _print_optional_output(
        console,
        title="Captured output:",
        output=result.captured_output.getvalue(),
    )
    _print_optional_output(
        console,
        title="Captured output from fixture teardowns:",
        output=result.fixture_teardown_output,
    )


def _print_test_failures(console: Console, failures: list[TestResult]) -> None:
    for result in failures:
        console.rule(f"{result.name}", style="bold red")
        failing_result = cast("FailedResult", result.result)

        _print_result_details(
            console,
            result=result,
            outcome=failing_result,
        )


def _print_test_errors(console: Console, errors: list[TestResult]) -> None:
    for result in errors:
        console.rule(f"{result.name}", style="bold dark_orange")
        error_result = cast("ErrorResult", result.result)

        _print_result_details(
            console,
            result=result,
            outcome=error_result,
        )


def _print_fixture_teardown_failures(
    console: Console, fixture_teardown_failures: list[TestResult]
) -> None:
    for result in fixture_teardown_failures:
        for teardown_failure in result.fixture_teardown_failures:
            console.rule(
                f"{result.name} - Fixture teardown: {teardown_failure.fixture_name}",
                style="bold red",
            )
            render_traceback(
                console,
                teardown_failure.exc_type,
                teardown_failure.exc_value,
                teardown_failure.traceback,
            )

        _print_optional_output(
            console,
            title="Captured output from fixture teardowns:",
            output=result.fixture_teardown_output,
        )


def _print_session_teardown_failures(
    console: Console, session_teardown_failures: list[TeardownFailure]
) -> None:
    for teardown_failure in session_teardown_failures:
        console.rule(
            f"Session fixture teardown: {teardown_failure.fixture_name}",
            style="bold red",
        )
        render_traceback(
            console,
            teardown_failure.exc_type,
            teardown_failure.exc_value,
            teardown_failure.traceback,
        )


def print_failures(
    console: Console,
    test_results: list[TestResult],
    session_teardown_failures: list[TeardownFailure] | None = None,
    session_teardown_output: str | None = None,
) -> None:
    """Print all test failures, fixture teardown failures, and session teardown failures."""
    if session_teardown_failures is None:
        session_teardown_failures = []

    groups = _collect_failure_groups(test_results)

    if (
        not groups.failures
        and not groups.errors
        and not groups.fixture_teardown_failures
        and not session_teardown_failures
    ):
        return

    console.print()
    console.rule("FAILURES", style="bold orange3", characters="=")
    console.print()

    _print_test_failures(console, groups.failures)
    _print_test_errors(console, groups.errors)
    _print_fixture_teardown_failures(console, groups.fixture_teardown_failures)
    _print_session_teardown_failures(console, session_teardown_failures)

    if session_teardown_output and (groups.failures or groups.errors):
        console.print()
        console.rule(
            "Output from session fixture teardowns",
            style="bold yellow",
        )
        console.print(session_teardown_output, markup=False, highlight=False)



import asyncio
import pdb  # noqa: T100
import sys
import time
from collections.abc import Callable
from inspect import isasyncgen, iscoroutine, isgenerator
from pathlib import Path
from types import TracebackType
from typing import cast

from snektest.annotations import Coroutine
from snektest.collection import TestsQueue
from snektest.fixtures import (
    get_active_function_fixtures,
    get_registered_session_fixtures,
    reset_session_fixtures,
)
from snektest.models import (
    AssertionFailure,
    BadRequestError,
    ErrorResult,
    FailedResult,
    PassedResult,
    TeardownFailure,
    TestName,
    TestResult,
    UnreachableError,
)
from snektest.output import maybe_capture_output
from snektest.presenter import print_failures, print_summary, print_test_result
from snektest.utils import get_test_function_markers, get_test_function_params


async def teardown_fixture(
    fixture_name: str,
    generator: object,
    *,
    exc_info_provider: Callable[
        [], tuple[object | None, object | None, TracebackType | None]
    ] = sys.exc_info,
) -> TeardownFailure | None:
    """Teardown a single fixture and return failure if it occurs."""
    try:
        if isasyncgen(generator):
            await anext(generator)
        elif isgenerator(generator):
            next(generator)
    except StopAsyncIteration, StopIteration:
        return None
    except Exception:
        exc_type, exc_value, traceback = exc_info_provider()
        if exc_type is None or exc_value is None or traceback is None:
            msg = "Invalid exception info gathered during teardown. This shouldn't be possible!"
            raise UnreachableError(msg) from None
        return TeardownFailure(
            fixture_name=fixture_name,
            exc_type=cast("type[BaseException]", exc_type),
            exc_value=cast("BaseException", exc_value),
            traceback=traceback,
        )
    else:
        msg = f"Incorrect fixture function {fixture_name} yielded more than once"
        raise BadRequestError(msg)


async def execute_test(
    name: TestName,
    func: Callable[..., Coroutine[None] | None],
    *,
    capture_output: bool = True,
    exc_info_provider: Callable[
        [], tuple[object | None, object | None, TracebackType | None]
    ] = sys.exc_info,
) -> TestResult:
    """Execute a single test function with fixtures and output capture."""
    with maybe_capture_output(capture_output) as (output_buffer, captured_warnings):
        param_values = ()
        if name.params_part:
            param_values = [
                param.value
                for param in get_test_function_params(func)[name.params_part]
            ]
        test_start = time.monotonic()
        try:
            res = func(*param_values)
            if iscoroutine(res):
                await res
            duration = time.monotonic() - test_start
            result = PassedResult()
        except (AssertionFailure, asyncio.CancelledError):
            duration = time.monotonic() - test_start
            exc_type, exc_value, traceback = exc_info_provider()
            if exc_type is None or exc_value is None or traceback is None:
                msg = "Invalid exception info gathered. This shouldn't be possible!"
                raise UnreachableError(msg) from None
            result = FailedResult(
                exc_type=cast("type[BaseException]", exc_type),
                exc_value=cast("BaseException", exc_value),
                traceback=traceback,
            )
        except Exception:
            duration = time.monotonic() - test_start
            exc_type, exc_value, traceback = exc_info_provider()
            if exc_type is None or exc_value is None or traceback is None:
                msg = "Invalid exception info gathered. This shouldn't be possible!"
                raise UnreachableError(msg) from None
            result = ErrorResult(
                exc_type=cast("type[BaseException]", exc_type),
                exc_value=cast("BaseException", exc_value),
                traceback=traceback,
            )

    with maybe_capture_output(capture_output) as (
        fixture_teardown_buffer,
        _,
    ):
        fixture_teardown_failures: list[TeardownFailure] = []
        for fixture_name, generator in get_active_function_fixtures():
            failure = await teardown_fixture(fixture_name, generator)
            if failure:
                fixture_teardown_failures.append(failure)

    fixture_teardown_output_value = fixture_teardown_buffer.getvalue() or None

    return TestResult(
        name=name,
        duration=duration,
        result=result,
        markers=get_test_function_markers(func),
        captured_output=output_buffer,
        fixture_teardown_failures=fixture_teardown_failures,
        fixture_teardown_output=fixture_teardown_output_value,
        warnings=captured_warnings,
    )


async def teardown_session_fixtures(
    *, capture_output: bool
) -> tuple[list[TeardownFailure], str | None]:
    """Teardown all session fixtures and return failures and output."""
    with maybe_capture_output(capture_output) as (teardown_output, _):
        session_teardown_failures: list[TeardownFailure] = []
        for fixture_func, (gen, _) in reversed(
            get_registered_session_fixtures().items()
        ):
            if gen is not None:
                fixture_name = fixture_func.co_name
                failure = await teardown_fixture(fixture_name, gen)
                if failure:
                    session_teardown_failures.append(failure)

    output_value = teardown_output.getvalue() or None
    return session_teardown_failures, output_value


def has_any_failures(
    test_results: list[TestResult], session_teardown_failures: list[TeardownFailure]
) -> tuple[bool, bool, bool]:
    """Check for test failures, fixture failures, and session failures."""
    has_test_failures = any(
        isinstance(result.result, (FailedResult, ErrorResult))
        for result in test_results
    )
    has_fixture_teardown_failures = any(
        result.fixture_teardown_failures for result in test_results
    )
    has_session_teardown_failures = len(session_teardown_failures) > 0
    return (
        has_test_failures,
        has_fixture_teardown_failures,
        has_session_teardown_failures,
    )


def _resolve_path(
    path: Path | None,
    *,
    resolver: Callable[[Path], Path] = Path.resolve,
) -> Path | None:
    if path is None:
        return None
    try:
        resolved = resolver(path)
    except FileNotFoundError:
        return path
    if resolved is path:
        return resolved
    if str(resolved):
        return resolved
    return resolved


def _trim_traceback(
    traceback: TracebackType, *, stop_at: TracebackType
) -> TracebackType:
    frames: list[TracebackType] = []
    current = traceback
    while current is not None:
        frames.append(current)
        if current is stop_at:
            break
        current = current.tb_next
    new_traceback: TracebackType | None = None
    for frame in reversed(frames):
        new_traceback = TracebackType(
            new_traceback, frame.tb_frame, frame.tb_lasti, frame.tb_lineno
        )
    if new_traceback is None:
        return traceback
    return new_traceback


def _traceback_for_file(
    traceback: TracebackType,
    *,
    preferred_file: Path | None,
    resolver: Callable[[Path], Path] = Path.resolve,
) -> TracebackType:
    preferred = _resolve_path(preferred_file, resolver=resolver)
    if preferred is None:
        return traceback

    selected: TracebackType | None = None
    current = traceback
    while current is not None:
        frame_path = Path(current.tb_frame.f_code.co_filename)
        resolved = _resolve_path(frame_path, resolver=resolver)
        if resolved == preferred:
            selected = current
        current = current.tb_next
    if selected is None:
        return traceback
    return _trim_traceback(traceback, stop_at=selected)


def _maybe_debug_test_result(
    test_result: TestResult,
    *,
    pdb_on_failure: bool,
    post_mortem: Callable[[TracebackType], None] = pdb.post_mortem,
    resolver: Callable[[Path], Path] = Path.resolve,
) -> bool:
    if not pdb_on_failure:
        return False
    if isinstance(test_result.result, (FailedResult, ErrorResult)):
        traceback = _traceback_for_file(
            test_result.result.traceback,
            preferred_file=test_result.name.file_path,
            resolver=resolver,
        )
        post_mortem(traceback)
        return True
    if test_result.fixture_teardown_failures:
        traceback = _traceback_for_file(
            test_result.fixture_teardown_failures[0].traceback,
            preferred_file=test_result.name.file_path,
            resolver=resolver,
        )
        post_mortem(traceback)
        return True
    return False


def _maybe_debug_session_teardown(
    session_teardown_failures: list[TeardownFailure],
    *,
    pdb_on_failure: bool,
    post_mortem: Callable[[TracebackType], None] = pdb.post_mortem,
) -> bool:
    if not pdb_on_failure or not session_teardown_failures:
        return False
    post_mortem(session_teardown_failures[0].traceback)
    return True


async def run_tests(
    queue: TestsQueue,
    *,
    capture_output: bool = True,
    pdb_on_failure: bool = False,
    post_mortem: Callable[[TracebackType], None] = pdb.post_mortem,
    resolver: Callable[[Path], Path] = Path.resolve,
) -> tuple[list[TestResult], list[TeardownFailure]]:
    """Run all tests from the queue and handle session fixture teardown."""
    total_duration = time.monotonic()
    test_results: list[TestResult] = []
    pdb_triggered = False
    try:
        while True:
            name, func = await queue.get()
            test_result = await execute_test(name, func, capture_output=capture_output)
            test_results.append(test_result)
            print_test_result(test_result)
            if not pdb_triggered and _maybe_debug_test_result(
                test_result,
                pdb_on_failure=pdb_on_failure,
                post_mortem=post_mortem,
                resolver=resolver,
            ):
                pdb_triggered = True
                break
    except asyncio.QueueShutDown:
        pass
    finally:
        session_teardown_failures, session_output = await teardown_session_fixtures(
            capture_output=capture_output
        )
        if not pdb_triggered and _maybe_debug_session_teardown(
            session_teardown_failures,
            pdb_on_failure=pdb_on_failure,
            post_mortem=post_mortem,
        ):
            pdb_triggered = True

        (
            has_test_failures,
            has_fixture_teardown_failures,
            has_session_teardown_failures,
        ) = has_any_failures(test_results, session_teardown_failures)

        session_output_for_display = None
        if session_output and (
            has_test_failures
            or has_fixture_teardown_failures
            or has_session_teardown_failures
        ):
            session_output_for_display = session_output

        print_failures(
            test_results,
            session_teardown_failures=session_teardown_failures,
            session_teardown_output=session_output_for_display,
        )
        print_summary(
            test_results,
            session_teardown_failures=session_teardown_failures,
            total_duration=time.monotonic() - total_duration,
        )
        reset_session_fixtures()
    return test_results, session_teardown_failures



import pathlib
from collections.abc import Callable
from types import TracebackType

from rich.console import Console
from rich.syntax import Syntax

import snektest


def render_traceback(  # noqa: PLR0913
    console: Console,
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: object,
    *,
    show_exception_line: bool = True,
    open_path: Callable[[str], list[str]] | None = None,
) -> None:
    """Render a traceback without a box, using Rich for syntax highlighting."""
    console.print("[bold]Traceback[/bold] [dim](most recent call last):[/dim]")

    tb = traceback
    snektest_path = str(snektest.__file__).rsplit("/", 1)[0]

    while tb:
        if not isinstance(tb, TracebackType):
            break

        frame = tb.tb_frame
        lineno = tb.tb_lineno
        filename = frame.f_code.co_filename
        name = frame.f_code.co_name

        if not filename.startswith(snektest_path):
            console.print(
                f'  File "[cyan]{filename}[/cyan]", line {lineno}, in [yellow]{name}[/yellow]'
            )

            try:
                if open_path is None:
                    with pathlib.Path(filename).open(encoding="utf-8") as f:
                        lines = f.readlines()
                else:
                    lines = open_path(filename)
                if 0 <= lineno - 1 < len(lines):
                    code_line = lines[lineno - 1].rstrip()
                    syntax = Syntax(
                        code_line,
                        "python",
                        theme="ansi_dark",
                        line_numbers=False,
                        padding=(0, 0, 0, 4),
                    )
                    console.print(syntax)
            except (OSError, IndexError):
                pass

        tb = tb.tb_next

    if not show_exception_line:
        return

    exc_name = exc_type.__name__
    exc_msg = str(exc_value)
    console.print(f"[red bold]{exc_name}[/red bold]: {exc_msg}")



from __future__ import annotations

from pathlib import Path

from pydantic import TypeAdapter, ValidationError

from snektest import assert_eq, assert_raises, test
from snektest.annotations import PyFilePath


@test()
def test_pyfilepath_json_schema_includes_format_and_string_type() -> None:
    schema = TypeAdapter(PyFilePath).json_schema()
    assert_eq(schema["type"], "string")
    assert_eq(schema["format"], "file-path")


@test()
def test_validate_pyfilepath_rejects_non_file() -> None:
    with assert_raises(ValidationError):
        TypeAdapter(PyFilePath).validate_python(Path())



from __future__ import annotations

"""Auto-start coverage in subprocesses.

When `COVERAGE_PROCESS_START` is set (e.g. to `pyproject.toml`), Coverage.py can
auto-enable measurement in child Python processes. This is critical for tests
that validate snektest via subprocesses.

See: https://coverage.readthedocs.io/en/latest/subprocess.html
"""

try:
    import coverage
except ModuleNotFoundError:
    coverage = None


def _maybe_start_coverage() -> None:
    if coverage is None:
        return

    _ = coverage.process_startup()


_maybe_start_coverage()



"""Tests for the --pdb CLI flag."""

import os
import re
import subprocess
from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file

TRACEBACK_WIDTH = 80


def _pad_traceback_line(line: str) -> str:
    padding = max(0, TRACEBACK_WIDTH - len(line))
    return f"{line}{' ' * padding}"


@test()
def test_pdb_stops_on_failure() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test
            from snektest.assertions import assert_eq

            @test()
            def test_failure() -> None:
                value = 1
                assert_eq(value, 2)
        """),
    )

    result = subprocess.run(
        ["uv", "run", "snektest", "--pdb", str(test_file)],
        input=b"p value\nc\n",
        capture_output=True,
        timeout=5,
        check=False,
        env={**os.environ, "RICH_WIDTH": "200"},
    )

    combined_output = result.stdout.decode() + result.stderr.decode()

    test_duration_match = re.search(r"FAIL \((\d+\.\d+)s\)", combined_output)
    summary_duration_match = re.search(
        r"1 failed, 0 passed in (\d+\.\d+)s", combined_output
    )
    assert test_duration_match is not None
    assert summary_duration_match is not None
    test_duration = test_duration_match.group(1)
    summary_duration = summary_duration_match.group(1)

    padded_line = _pad_traceback_line("        assert_eq(value, 2)")
    expected_output = dedent(f"""
        {test_file}::test_failure ... FAIL ({test_duration}s)
        > {test_file}(8)test_failure()
        -> assert_eq(value, 2)
        (Pdb) 1
        (Pdb)\u0020
        =================================== FAILURES ===================================

        ─────────────── {test_file}::test_failure ───────────────
        Traceback (most recent call last):
          File "{test_file}", line 8, in test_failure
        {padded_line}
        E       1 != 2
        ─────────────────────────────────── SUMMARY ────────────────────────────────────
        FAILED {test_file}::test_failure - 1 != 2

        ───────────────────────── 1 failed, 0 passed in {summary_duration}s ──────────────────────────
        """).lstrip()

    assert_eq(result.returncode, 1)
    assert_eq(combined_output, expected_output)


@test()
def test_pdb_stops_on_fixture_teardown_failure() -> None:
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import load_fixture, test

            def fix():
                value = "fixture value"
                yield value
                raise RuntimeError("fixture teardown failed")

            @test()
            def test_fix() -> None:
                _ = load_fixture(fix())
        """),
    )

    result = subprocess.run(
        ["uv", "run", "snektest", "--pdb", str(test_file)],
        input=b"p value\nc\n",
        capture_output=True,
        timeout=5,
        check=False,
        env={**os.environ, "RICH_WIDTH": "200"},
    )

    combined_output = result.stdout.decode() + result.stderr.decode()

    test_duration_match = re.search(r"OK \((\d+\.\d+)s\)", combined_output)
    summary_duration_match = re.search(
        r"1 fixture teardown failed, 1 passed in (\d+\.\d+)s", combined_output
    )
    assert test_duration_match is not None
    assert summary_duration_match is not None
    test_duration = test_duration_match.group(1)
    summary_duration = summary_duration_match.group(1)

    padded_line = _pad_traceback_line(
        '        raise RuntimeError("fixture teardown failed")'
    )
    expected_output = dedent(f"""
        {test_file}::test_fix ... OK ({test_duration}s)
        > {test_file}(7)fix()
        -> raise RuntimeError("fixture teardown failed")
        (Pdb) 'fixture value'
        (Pdb)\u0020
        =================================== FAILURES ===================================

        ───── {test_file}::test_fix - Fixture teardown: fix ─────
        Traceback (most recent call last):
          File "{test_file}", line 7, in fix
        {padded_line}
        RuntimeError: fixture teardown failed
        ─────────────────────────────────── SUMMARY ────────────────────────────────────
        FIXTURE TEARDOWN FAILED {test_file}::test_fix - fix: fixture teardown failed

        ───────────────── 1 fixture teardown failed, 1 passed in {summary_duration}s ─────────────────
        """).lstrip()

    assert_eq(result.returncode, 1)
    assert_eq(combined_output, expected_output)



from __future__ import annotations

from io import StringIO
from pathlib import Path
from types import TracebackType

from rich.console import Console

from snektest import assert_in, test
from snektest.models import (
    ErrorResult,
    FailedResult,
    PassedResult,
    TestName,
    TestResult,
)
from snektest.presenter.errors import print_failures
from snektest.presenter.summary import print_summary
from snektest.presenter.traceback import render_traceback


def _traceback_from_exception(exc: BaseException) -> TracebackType:
    try:
        raise exc
    except type(exc) as e:
        tb = e.__traceback__
        assert tb is not None
        return tb


@test()
def test_print_failures_includes_captured_and_fixture_teardown_output() -> None:
    console = Console(record=True)
    tb = _traceback_from_exception(RuntimeError("boom"))

    failing = TestResult(
        name=TestName(file_path=Path("x.py"), func_name="t", params_part=""),
        duration=0.0,
        result=FailedResult(
            exc_type=RuntimeError, exc_value=RuntimeError("boom"), traceback=tb
        ),
        markers=(),
        captured_output=StringIO("captured"),
        fixture_teardown_failures=[],
        fixture_teardown_output="fixture-teardown-output",
        warnings=[],
    )

    print_failures(
        console,
        [failing],
        session_teardown_failures=None,
        session_teardown_output="session-out",
    )
    print_failures(
        console,
        [failing],
        session_teardown_failures=None,
        session_teardown_output=None,
    )

    text = console.export_text()
    assert_in("Captured output:", text)
    assert_in("Captured output from fixture teardowns:", text)
    assert_in("Output from session fixture teardowns", text)


@test()
def test_print_summary_error_without_message() -> None:
    console = Console(record=True)
    tb = _traceback_from_exception(RuntimeError("x"))

    result = TestResult(
        name=TestName(file_path=Path("x.py"), func_name="e", params_part=""),
        duration=0.0,
        result=ErrorResult(
            exc_type=RuntimeError, exc_value=RuntimeError(""), traceback=tb
        ),
        markers=(),
        captured_output=StringIO(""),
        fixture_teardown_failures=[],
        fixture_teardown_output=None,
        warnings=[],
    )

    print_summary(console, [result], 0.0, session_teardown_failures=[])
    print_summary(console, [result], 0.0, session_teardown_failures=[])


@test()
def test_print_summary_warnings_and_failed_without_message() -> None:
    console = Console(record=True)

    passed = TestResult(
        name=TestName(file_path=Path("x.py"), func_name="p", params_part=""),
        duration=0.0,
        result=PassedResult(),
        markers=(),
        captured_output=StringIO(""),
        fixture_teardown_failures=[],
        fixture_teardown_output=None,
        warnings=["warning"],
    )

    tb = _traceback_from_exception(RuntimeError("x"))
    failed = TestResult(
        name=TestName(file_path=Path("x.py"), func_name="f", params_part=""),
        duration=0.0,
        result=FailedResult(
            exc_type=RuntimeError, exc_value=RuntimeError(""), traceback=tb
        ),
        markers=(),
        captured_output=StringIO(""),
        fixture_teardown_failures=[],
        fixture_teardown_output=None,
        warnings=[],
    )

    print_summary(console, [passed, failed], 0.0, session_teardown_failures=None)
    print_summary(console, [passed, failed], 0.0, session_teardown_failures=None)
    text = console.export_text()
    assert_in("WARNINGS", text)
    assert_in("SUMMARY", text)


@test()
def test_render_traceback_handles_non_traceback_object() -> None:
    console = Console(record=True)

    render_traceback(
        console,
        RuntimeError,
        RuntimeError("x"),
        object(),
        show_exception_line=False,
    )


@test()
def test_render_traceback_handles_oserror_open() -> None:
    console = Console(record=True)
    tb = _traceback_from_exception(RuntimeError("x"))

    def open_path(_: str) -> list[str]:
        raise OSError

    render_traceback(console, RuntimeError, RuntimeError("x"), tb, open_path=open_path)
    render_traceback(console, RuntimeError, RuntimeError("x"), tb)
    render_traceback(console, RuntimeError, RuntimeError("x"), tb)
    render_traceback(console, RuntimeError, RuntimeError("x"), tb)
    render_traceback(console, RuntimeError, RuntimeError("x"), tb)



from collections.abc import AsyncGenerator, Generator
from inspect import isasyncgen, isgenerator
from types import CodeType
from typing import Any

from snektest.annotations import Coroutine
from snektest.models import UnreachableError
from snektest.utils import get_code_from_generator, get_func_name_from_generator

_SESSION_FIXTURES: dict[
    CodeType, tuple[AsyncGenerator[Any] | Generator[Any] | None, object]
] = {}
_FUNCTION_FIXTURES: list[AsyncGenerator[Any] | Generator[Any]] = []


def register_session_fixture(
    fixture_code: CodeType,
) -> None:
    """Register a session-scoped fixture."""
    if fixture_code not in _SESSION_FIXTURES:
        _SESSION_FIXTURES[fixture_code] = (None, None)


def get_registered_session_fixtures() -> dict[
    CodeType, tuple[AsyncGenerator[Any] | Generator[Any] | None, object]
]:
    """Get all registered session fixtures."""
    return _SESSION_FIXTURES


def reset_session_fixtures() -> None:
    """Clear cached session fixtures for a fresh test run."""
    _SESSION_FIXTURES.clear()


def is_session_fixture(fixture_code: CodeType) -> bool:
    """Check if a fixture code object is registered as a session fixture."""
    return fixture_code in _SESSION_FIXTURES


def load_session_fixture[R](fixture_gen: AsyncGenerator[R] | Generator[R]) -> R:
    """Load a session-scoped fixture, creating it on first use and reusing thereafter."""
    fixture_code = get_code_from_generator(fixture_gen)
    try:
        gen, result = _SESSION_FIXTURES[fixture_code]
        if gen is None:
            gen = fixture_gen
            if isasyncgen(gen):

                async def result_updater() -> R:
                    gen, _ = _SESSION_FIXTURES[fixture_code]
                    if not isasyncgen(gen):
                        msg = "This should not happen I think"
                        raise UnreachableError(msg)
                    result = await anext(gen)

                    async def async_wrapper() -> R:
                        return result

                    _SESSION_FIXTURES[fixture_code] = (gen, async_wrapper())
                    return result

                result = result_updater()
            elif isgenerator(gen):
                result = next(gen)
            _SESSION_FIXTURES[fixture_code] = (gen, result)
    except KeyError:
        msg = f"Function {fixture_code.co_qualname} was not registered as a session fixture. This shouldn't be possible!"
        raise UnreachableError(msg) from None
    else:
        return result  # pyright: ignore[reportReturnType]


def load_function_fixture[R](
    fixture_gen: AsyncGenerator[R] | Generator[R],
) -> Coroutine[R] | R:
    """Load a function-scoped fixture by registering and yielding its value."""
    if isasyncgen(fixture_gen):
        _FUNCTION_FIXTURES.append(fixture_gen)
        return anext(fixture_gen)
    if isgenerator(fixture_gen):
        _FUNCTION_FIXTURES.append(fixture_gen)
        return next(fixture_gen)
    msg = "Fixture must be a generator or async generator"
    raise UnreachableError(msg)


def get_active_function_fixtures() -> list[
    tuple[str, AsyncGenerator[Any] | Generator[Any]]
]:
    """Return the list of active function fixtures, as (function_name, generator) tuples.

    Returns:
        List of (fixture_name, generator) tuples in reverse order.
    """
    fixtures_to_teardown: list[tuple[str, AsyncGenerator[Any] | Generator[Any]]] = []
    # Returning active fixtures in reverse order makes setup/teardown first-in-last-out
    for generator in reversed(_FUNCTION_FIXTURES):
        fixture_name = get_func_name_from_generator(generator)
        fixtures_to_teardown.append((fixture_name, generator))

    _FUNCTION_FIXTURES.clear()
    return fixtures_to_teardown



"""Meta tests for async fixtures and parameterization."""

from textwrap import dedent

from snektest import load_fixture, test
from snektest.assertions import assert_eq
from testutils.fixtures import tmp_dir_fixture
from testutils.helpers import create_test_file, run_test_subprocess


@test()
def test_async_session_fixture() -> None:
    """Test async session fixture works and is reused across tests."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import AsyncGenerator
            from snektest import load_fixture, session_fixture, test, assert_eq

            @session_fixture()
            async def fixture_for_session() -> AsyncGenerator[int]:
                print("session fixture started")
                yield 10
                print("session fixture ended")

            @test()
            async def test_with_session_fixture() -> None:
                session_fixture_result = await load_fixture(fixture_for_session())
                assert_eq(session_fixture_result, 10)

            @test()
            async def another_test_with_session_fixture() -> None:
                session_fixture_result = await load_fixture(fixture_for_session())
                assert_eq(session_fixture_result, 10)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 2)
    assert_eq(result["failed"], 0)


@test()
def test_async_function_fixture() -> None:
    """Test async function fixture works."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import AsyncGenerator
            from snektest import load_fixture, test, assert_eq

            async def simple_fixture() -> AsyncGenerator[str]:
                print("simple async fixture started")
                yield "some fixture"
                print("simple async fixture ended")

            @test()
            async def test_with_simple_fixture() -> None:
                fixture = await load_fixture(simple_fixture())
                assert_eq(fixture, "some fixture")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)


@test()
def test_parameterized_async_test() -> None:
    """Test parameterized async tests work."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_eq
            from snektest.models import Param

            first_param_set = [
                Param(" bab ", "spaces both sides"),
                Param(" bab", "space left side"),
                Param("bab ", "space right side"),
                Param("bab", "no spaces"),
            ]

            @test(first_param_set)
            async def test_1_params(param1: str) -> None:
                assert_eq(param1.strip(), "bab")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 4)
    assert_eq(result["failed"], 0)


@test()
def test_multi_param_async_test() -> None:
    """Test async tests with multiple parameter sets."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from snektest import test, assert_eq, assert_in
            from snektest.models import Param

            first_param_set = [
                Param(" bab ", "spaces both sides"),
                Param("bab", "no spaces"),
            ]
            second_param_set = [
                Param(5, "attempt 1"),
                Param(10, "attempt 2"),
            ]

            @test(first_param_set, second_param_set)
            async def test_2_params(param1: str, param2: int) -> None:
                assert_eq(param1.strip(), "bab")
                assert_in(param2, [5, 10])
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 4)
    assert_eq(result["failed"], 0)


@test()
def test_async_fixture_with_params() -> None:
    """Test async fixture that accepts parameters."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import AsyncGenerator
            from snektest import load_fixture, test, assert_eq
            from snektest.models import Param

            async def fixture_with_param(param1: str) -> AsyncGenerator[str]:
                yield param1

            @test([Param("value1", "param1"), Param("value2", "param2")])
            async def test_with_param_fixture(param1: str) -> None:
                result = await load_fixture(fixture_with_param(param1))
                assert_eq(result, param1)
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 2)
    assert_eq(result["failed"], 0)


@test()
def test_async_fixture_teardown() -> None:
    """Test async fixture teardown is executed."""
    tmp_dir = load_fixture(tmp_dir_fixture())

    test_file = create_test_file(
        tmp_dir,
        dedent("""
            from collections.abc import AsyncGenerator
            from snektest import load_fixture, test, assert_eq
            from snektest.models import Param

            async def fixture_with_teardown_and_param(param: str) -> AsyncGenerator[str]:
                print(f"setup: {param}")
                yield param
                print(f"teardown: {param}")

            @test([Param("the number", "single-param")])
            async def test_with_param_and_teardown_fixture(param1: str) -> None:
                fixture_result = await load_fixture(fixture_with_teardown_and_param(param1))
                assert_eq(param1, "the number")
        """),
    )

    result = run_test_subprocess(test_file)
    assert_eq(result["passed"], 1)
    assert_eq(result["failed"], 0)



"""Helper functions for testing snektest itself."""

import json
import subprocess
import sys
from pathlib import Path
from typing import Any


def create_test_file(
    tmp_dir: Path, content: str, *, name: str = "test_generated"
) -> Path:
    """Create a temporary Python test file.

    Args:
        tmp_dir: Directory to create file in
        content: Python code as string (use textwrap.dedent!)
        name: Filename without .py extension

    Returns:
        Path to created file
    """
    filepath = tmp_dir / f"{name}.py"
    _ = filepath.write_text(content)
    return filepath


def run_test_subprocess(test_file: Path) -> dict[str, Any]:
    """Run snektest subprocess on test file and return structured results.

    Args:
        test_file: Path to test file to run

    Returns:
        Dict with keys: passed, failed, fixture_teardown_failed,
                       session_teardown_failed, returncode
    """
    cmd = [sys.executable, "-m", "snektest.cli", "--json-output", str(test_file)]

    result = subprocess.run(
        cmd,
        check=False,
        capture_output=True,
        text=True,
        timeout=0.5,
    )

    lines = result.stdout.strip().split("\n")
    json_line = lines[-1]
    results = json.loads(json_line)
    results["returncode"] = result.returncode
    return results



"""Unit tests for basic snektest functionality."""

from pathlib import Path

from snektest import Param, test
from snektest.assertions import (
    assert_eq,
    assert_false,
    assert_ge,
    assert_gt,
    assert_in,
    assert_is,
    assert_is_none,
    assert_is_not,
    assert_is_not_none,
    assert_isinstance,
    assert_le,
    assert_len,
    assert_lt,
    assert_ne,
    assert_not_in,
    assert_not_isinstance,
    assert_raises,
    assert_true,
    fail,
)
from snektest.models import AssertionFailure, TestName
from snektest.utils import (
    get_test_function_params,
    is_test_function,
    mark_test_function,
)


@test()
def test_mark_test_function_basic() -> None:
    """Test that mark_test_function properly marks a function."""

    def dummy_func() -> None:
        pass

    mark_test_function(dummy_func, (), ())
    assert_true(is_test_function(dummy_func))


@test()
def test_is_test_function_unmarked() -> None:
    """Test that unmarked functions return False."""

    def dummy_func() -> None:
        pass

    assert_false(is_test_function(dummy_func))


@test()
def test_get_test_function_params_no_params() -> None:
    """Test getting params from a test with no parameters."""

    def dummy_func() -> None:
        pass

    mark_test_function(dummy_func, (), ())
    result = get_test_function_params(dummy_func)
    assert_eq(result, {"": ()})


@test()
def test_get_test_function_params_single_param_list() -> None:
    """Test getting params from a test with one parameter list."""

    def dummy_func(x: int) -> None:
        pass

    params = ([Param(value=1, name="1"), Param(value=2, name="2")],)
    mark_test_function(dummy_func, params, ())

    result = get_test_function_params(dummy_func)
    assert_eq(
        result,
        {
            "1": (params[0][0],),
            "2": (params[0][1],),
        },
    )


@test()
def test_param_to_dict_empty() -> None:
    """Test Param.to_dict with no params."""
    result = Param.to_dict(())
    assert_eq(result, {"": ()})


@test()
def test_param_to_dict_single_list() -> None:
    """Test Param.to_dict with a single param list."""
    params = ([Param(value=1, name="one"), Param(value=2, name="two")],)
    result = Param.to_dict(params)

    assert_eq(
        result,
        {
            "one": (params[0][0],),
            "two": (params[0][1],),
        },
    )


@test()
def test_param_to_dict_multiple_lists() -> None:
    """Test Param.to_dict creates cartesian product of param lists."""
    params = (
        [Param(value=1, name="a"), Param(value=2, name="b")],
        [Param(value="x", name="X"), Param(value="y", name="Y")],
    )
    result = Param.to_dict(params)

    assert_eq(
        result,
        {
            "a, X": (params[0][0], params[1][0]),
            "a, Y": (params[0][0], params[1][1]),
            "b, X": (params[0][1], params[1][0]),
            "b, Y": (params[0][1], params[1][1]),
        },
    )


@test()
def test_testname_str_no_params() -> None:
    """Test TestName string representation without params."""
    name = TestName(
        file_path=Path("tests/test_foo.py"), func_name="test_bar", params_part=""
    )
    assert_eq(str(name), "tests/test_foo.py::test_bar")


@test()
def test_testname_str_with_params() -> None:
    """Test TestName string representation with params."""
    name = TestName(
        file_path=Path("tests/test_foo.py"),
        func_name="test_bar",
        params_part="x=1, y=2",
    )
    assert_eq(str(name), "tests/test_foo.py::test_bar[x=1, y=2]")


@test()
def test_assert_eq_passes() -> None:
    """Test assert_eq with equal values."""
    assert_eq(1, 1)
    assert_eq("hello", "hello")
    assert_eq([1, 2], [1, 2])


@test()
def test_assert_eq_fails() -> None:
    """Test assert_eq raises AssertionFailure on inequality."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(1, 2)

    assert_eq(exc_info.exception.actual, 1)
    assert_eq(exc_info.exception.expected, 2)
    assert_eq(exc_info.exception.operator, "==")


@test()
def test_assert_ne_passes() -> None:
    """Test assert_ne with unequal values."""
    assert_ne(1, 2)
    assert_ne("hello", "world")


@test()
def test_assert_ne_fails() -> None:
    """Test assert_ne raises AssertionFailure on equality."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_ne(5, 5)

    assert_eq(exc_info.exception.actual, 5)
    assert_eq(exc_info.exception.expected, 5)
    assert_eq(exc_info.exception.operator, "!=")


@test()
def test_assert_true_passes() -> None:
    """Test assert_true with True value."""
    assert_true(True)


@test()
def test_assert_true_fails_with_truthy() -> None:
    """Test assert_true fails even with truthy values (identity check)."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_true(1)

    assert_eq(exc_info.exception.actual, 1)
    assert_eq(exc_info.exception.expected, True)


@test()
def test_assert_false_passes() -> None:
    """Test assert_false with False value."""
    assert_false(False)


@test()
def test_assert_false_fails_with_falsy() -> None:
    """Test assert_false fails even with falsy values (identity check)."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_false(0)

    assert_eq(exc_info.exception.actual, 0)
    assert_eq(exc_info.exception.expected, False)


@test()
def test_assert_is_none_passes() -> None:
    """Test assert_is_none with None."""
    assert_is_none(None)


@test()
def test_assert_is_none_fails() -> None:
    """Test assert_is_none fails with non-None values."""
    with assert_raises(AssertionFailure):
        assert_is_none(0)


@test()
def test_assert_is_not_none_passes() -> None:
    """Test assert_is_not_none with non-None values."""
    _ = assert_is_not_none(0)
    _ = assert_is_not_none("")
    _ = assert_is_not_none(False)


@test()
def test_assert_is_not_none_fails() -> None:
    """Test assert_is_not_none fails with None."""
    with assert_raises(AssertionFailure):
        _ = assert_is_not_none(None)  # pyright: ignore[reportUnknownVariableType]


@test()
def test_assert_is_passes() -> None:
    """Test assert_is with same object."""
    obj: list[int] = []
    assert_is(obj, obj)


@test()
def test_assert_is_fails() -> None:
    """Test assert_is fails with different objects."""
    with assert_raises(AssertionFailure):
        assert_is([], [])


@test()
def test_assert_is_not_passes() -> None:
    """Test assert_is_not with different objects."""
    assert_is_not([], [])


@test()
def test_assert_is_not_fails() -> None:
    """Test assert_is_not fails with same object."""
    obj: list[int] = []
    with assert_raises(AssertionFailure):
        assert_is_not(obj, obj)


@test()
def test_assert_lt_passes() -> None:
    """Test assert_lt with less than values."""
    assert_lt(1, 2)
    assert_lt("a", "b")


@test()
def test_assert_lt_fails() -> None:
    """Test assert_lt fails when not less than."""
    with assert_raises(AssertionFailure):
        assert_lt(2, 1)


@test()
def test_assert_gt_passes() -> None:
    """Test assert_gt with greater than values."""
    assert_gt(2, 1)
    assert_gt("b", "a")


@test()
def test_assert_gt_fails() -> None:
    """Test assert_gt fails when not greater than."""
    with assert_raises(AssertionFailure):
        assert_gt(1, 2)


@test()
def test_assert_le_passes() -> None:
    """Test assert_le with less than or equal values."""
    assert_le(1, 2)
    assert_le(2, 2)


@test()
def test_assert_le_fails() -> None:
    """Test assert_le fails when greater than."""
    with assert_raises(AssertionFailure):
        assert_le(3, 2)


@test()
def test_assert_ge_passes() -> None:
    """Test assert_ge with greater than or equal values."""
    assert_ge(2, 1)
    assert_ge(2, 2)


@test()
def test_assert_ge_fails() -> None:
    """Test assert_ge fails when less than."""
    with assert_raises(AssertionFailure):
        assert_ge(1, 2)


@test()
def test_assert_in_passes() -> None:
    """Test assert_in with member in container."""
    assert_in(1, [1, 2, 3])
    assert_in("a", "abc")
    assert_in("key", {"key": "value"})


@test()
def test_assert_in_fails() -> None:
    """Test assert_in fails when member not in container."""
    with assert_raises(AssertionFailure):
        assert_in(4, [1, 2, 3])


@test()
def test_assert_not_in_passes() -> None:
    """Test assert_not_in with member not in container."""
    assert_not_in(4, [1, 2, 3])
    assert_not_in("d", "abc")


@test()
def test_assert_not_in_fails() -> None:
    """Test assert_not_in fails when member in container."""
    with assert_raises(AssertionFailure):
        assert_not_in(1, [1, 2, 3])


@test()
def test_assert_isinstance_passes() -> None:
    """Test assert_isinstance with correct type."""
    assert_isinstance(1, int)
    assert_isinstance("hello", str)
    assert_isinstance([1, 2], list)


@test()
def test_assert_isinstance_fails() -> None:
    """Test assert_isinstance fails with wrong type."""
    with assert_raises(AssertionFailure):
        assert_isinstance(1, str)


@test()
def test_assert_isinstance_multiple_types() -> None:
    """Test assert_isinstance with tuple of types."""
    assert_isinstance(1, (int, str))
    assert_isinstance("hello", (int, str))


@test()
def test_assert_not_isinstance_passes() -> None:
    """Test assert_not_isinstance with wrong type."""
    assert_not_isinstance(1, str)
    assert_not_isinstance("hello", int)


@test()
def test_assert_not_isinstance_fails() -> None:
    """Test assert_not_isinstance fails with correct type."""
    with assert_raises(AssertionFailure):
        assert_not_isinstance(1, int)


@test()
def test_assert_len_passes() -> None:
    """Test assert_len with correct length."""
    assert_len([1, 2, 3], 3)
    assert_len("hello", 5)
    assert_len({"a": 1, "b": 2}, 2)


@test()
def test_assert_len_fails() -> None:
    """Test assert_len fails with wrong length."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_len([1, 2, 3], 5)

    assert_eq(exc_info.exception.actual, 3)
    assert_eq(exc_info.exception.expected, 5)


@test()
def test_fail_always_raises() -> None:
    """Test that fail always raises AssertionFailure."""
    with assert_raises(AssertionFailure):
        fail()


@test()
def test_fail_with_message() -> None:
    """Test fail with custom message."""
    with assert_raises(AssertionFailure) as exc_info:
        fail("Custom failure message")

    assert_eq(str(exc_info.exception), "Custom failure message")


@test()
def test_assertion_failure_custom_message() -> None:
    """Test that custom messages work in assertions."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(1, 2, msg="Values don't match")

    assert_eq(str(exc_info.exception), "Values don't match")



from __future__ import annotations

import asyncio
import tempfile
from pathlib import Path
from typing import cast

from pydantic import TypeAdapter

from snektest import assert_eq, assert_raises, test
from snektest.annotations import PyFilePath
from snektest.collection import TestsQueue, load_tests_from_file
from snektest.models import CollectionError, FilterItem


@test()
async def test_load_tests_from_file_caches_module() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        tmp_dir = Path(tmp)
        test_file = tmp_dir / "test_collection_generated.py"
        _ = test_file.write_text(
            """
from snektest import test

@test()
def test_one() -> None:
    pass
""".lstrip()
        )

        file_path = cast(
            "PyFilePath", TypeAdapter(PyFilePath).validate_python(test_file)
        )
        filter_item = FilterItem(str(test_file))
        loop = asyncio.get_running_loop()

        queue: TestsQueue = TestsQueue()
        load_tests_from_file(file_path, filter_item, queue, loop, mark=None)
        _ = await asyncio.wait_for(queue.get(), timeout=1)

        queue2: TestsQueue = TestsQueue()
        load_tests_from_file(file_path, filter_item, queue2, loop, mark=None)
        _ = await asyncio.wait_for(queue2.get(), timeout=1)


@test()
async def test_load_tests_from_file_filters_function_and_params() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        tmp_dir = Path(tmp)

        test_file = tmp_dir / "test_collection_generated_params.py"
        _ = test_file.write_text(
            """
from snektest import test
from snektest.models import Param

@test([Param(1, 'one')])
def test_param(x: int) -> None:
    _ = x

@test()
def test_other() -> None:
    pass
""".lstrip()
        )

        file_path = cast(
            "PyFilePath", TypeAdapter(PyFilePath).validate_python(test_file)
        )
        loop = asyncio.get_running_loop()

        queue: TestsQueue = TestsQueue()
        load_tests_from_file(
            file_path,
            FilterItem(f"{test_file}::test_other"),
            queue,
            loop,
            mark=None,
        )
        name, _func = await asyncio.wait_for(queue.get(), timeout=1)
        assert_eq(name.func_name, "test_other")

        queue2: TestsQueue = TestsQueue()
        load_tests_from_file(
            file_path,
            FilterItem(f"{test_file}::test_param[does not match]"),
            queue2,
            loop,
            mark=None,
        )
        queue2.shutdown()
        with assert_raises(asyncio.QueueShutDown):
            _ = await queue2.get()


@test()
def test_load_tests_from_file_spec_loader_failure_raises_collection_error() -> None:
    def fake_spec(_name: object, _path: object) -> None:
        return None

    with assert_raises(CollectionError):
        queue: TestsQueue = TestsQueue()
        loop = asyncio.new_event_loop()
        try:
            load_tests_from_file(
                cast(
                    "PyFilePath",
                    TypeAdapter(PyFilePath).validate_python(Path(__file__)),
                ),
                FilterItem(str(Path(__file__))),
                queue,
                loop,
                mark=None,
                spec_loader=fake_spec,
            )
        finally:
            loop.close()



from __future__ import annotations

from dataclasses import dataclass

from rich.console import Console

from snektest.models import (
    ErrorResult,
    FailedResult,
    PassedResult,
    TeardownFailure,
    TestResult,
)


@dataclass(frozen=True)
class RunCounts:
    passed: int
    failed: int
    errors: int
    fixture_teardown_failed: int
    session_teardown_failed: int


def _print_warnings(console: Console, test_results: list[TestResult]) -> None:
    """Print warnings from test results."""
    all_warnings = [w for result in test_results for w in result.warnings]
    if all_warnings:
        console.print()
        console.rule("WARNINGS", style="bold yellow")
        for warning in all_warnings:
            console.print(warning, markup=False, style="yellow")
        console.print()


def _print_test_failures(console: Console, test_results: list[TestResult]) -> None:
    """Print test failures."""
    for result in test_results:
        if (failed_result := result.result) and isinstance(failed_result, FailedResult):
            error_msg = str(failed_result.exc_value)
            console.print("FAILED", style="red", end=" ")
            if error_msg:
                console.print(f"{result.name} - {error_msg}", markup=False)
            else:
                console.print(f"{result.name}", markup=False)


def _print_test_errors(console: Console, test_results: list[TestResult]) -> None:
    """Print test errors (unexpected exceptions)."""
    for result in test_results:
        if (error_result := result.result) and isinstance(error_result, ErrorResult):
            error_msg = str(error_result.exc_value)
            console.print("ERROR", style="dark_orange", end=" ")
            if error_msg:
                console.print(f"{result.name} - {error_msg}", markup=False)
            else:
                console.print(f"{result.name}", markup=False)


def _print_fixture_teardown_failures(
    console: Console, test_results: list[TestResult]
) -> None:
    """Print fixture teardown failures."""
    for result in test_results:
        for teardown_failure in result.fixture_teardown_failures:
            console.print("FIXTURE TEARDOWN FAILED", style="red", end=" ")
            console.print(
                f"{result.name} - {teardown_failure.fixture_name}: {teardown_failure.exc_value}",
                markup=False,
            )


def _print_session_teardown_failures(
    console: Console, session_teardown_failures: list[TeardownFailure]
) -> None:
    """Print session teardown failures."""
    for teardown_failure in session_teardown_failures:
        console.print("SESSION FIXTURE TEARDOWN FAILED", style="red", end=" ")
        console.print(
            f"{teardown_failure.fixture_name}: {teardown_failure.exc_value}",
            markup=False,
        )


def _has_failures(counts: RunCounts) -> bool:
    return (
        counts.failed > 0
        or counts.errors > 0
        or counts.fixture_teardown_failed > 0
        or counts.session_teardown_failed > 0
    )


def _build_status_text(*, counts: RunCounts, total_duration: float) -> tuple[str, str]:
    """Build status text and set its color."""
    status_color = "red" if _has_failures(counts) else "green"
    status_text = ""
    if counts.failed > 0:
        status_text += f"{counts.failed} failed, "
    if counts.errors > 0:
        status_text += f"{counts.errors} error, "
    if counts.fixture_teardown_failed > 0:
        status_text += f"{counts.fixture_teardown_failed} fixture teardown failed, "
    if counts.session_teardown_failed > 0:
        status_text += (
            f"{counts.session_teardown_failed} session fixture teardown failed, "
        )
    status_text += f"{counts.passed} passed in {total_duration:.2f}s"
    return status_text, f"bold {status_color}"


def print_summary(
    console: Console,
    test_results: list[TestResult],
    total_duration: float,
    session_teardown_failures: list[TeardownFailure] | None = None,
) -> None:
    """Print test summary with counts and status."""
    if session_teardown_failures is None:
        session_teardown_failures = []

    passed = sum(
        1 for result in test_results if isinstance(result.result, PassedResult)
    )
    failed = sum(
        1 for result in test_results if isinstance(result.result, FailedResult)
    )
    errors = sum(1 for result in test_results if isinstance(result.result, ErrorResult))
    fixture_teardown_failed = sum(
        len(result.fixture_teardown_failures) for result in test_results
    )
    session_teardown_failed = len(session_teardown_failures)

    counts = RunCounts(
        passed=passed,
        failed=failed,
        errors=errors,
        fixture_teardown_failed=fixture_teardown_failed,
        session_teardown_failed=session_teardown_failed,
    )

    _print_warnings(console, test_results)

    if _has_failures(counts):
        console.rule("SUMMARY", style="wheat1")
        _print_test_failures(console, test_results)
        _print_test_errors(console, test_results)
        _print_fixture_teardown_failures(console, test_results)
        _print_session_teardown_failures(console, session_teardown_failures)
        console.print()

    status_text, status_style = _build_status_text(
        counts=counts,
        total_duration=total_duration,
    )
    console.rule(status_text, style=status_style)



import asyncio
from collections.abc import AsyncGenerator, Callable, Generator
from concurrent.futures import Future
from enum import Enum
from functools import wraps
from inspect import Parameter, Signature, iscoroutinefunction
from typing import Any, Protocol, TypeVar, cast, overload

from hypothesis import given

from snektest.annotations import Coroutine
from snektest.fixtures import (
    is_session_fixture,
    load_function_fixture,
    load_session_fixture,
    register_session_fixture,
)
from snektest.models import Param
from snektest.utils import get_code_from_generator, mark_test_function

_given = cast("Any", given)

T_co = TypeVar("T_co", covariant=True)


class SearchStrategy(Protocol[T_co]):
    def example(self) -> T_co: ...


class Marker(Enum):
    FAST = "fast"
    MEDIUM = "medium"
    SLOW = "slow"


def _normalize_marker_entry(entry: object) -> str:
    if isinstance(entry, Marker):
        return entry.value
    msg = "Markers must be Marker values"
    raise TypeError(msg)


def _normalize_marker_sequence(
    markers: list[object] | tuple[object, ...],
) -> tuple[str, ...]:
    return tuple(_normalize_marker_entry(entry) for entry in markers)


def _normalize_markers(mark: object | None) -> tuple[str, ...]:
    if mark is None:
        return ()
    if isinstance(mark, Marker):
        return (_normalize_marker_entry(mark),)
    if isinstance(mark, list):
        return _normalize_marker_sequence(cast("list[object]", mark))
    if isinstance(mark, tuple):
        return _normalize_marker_sequence(cast("tuple[object, ...]", mark))
    msg = "Markers must be a Marker or list/tuple of Marker values"
    raise TypeError(msg)


@overload
def test(
    *params: list[Param[Any]],
    mark: Marker | list[Marker] | tuple[Marker, ...] | None = None,
) -> Callable[
    [Callable[[*tuple[Any, ...]], Coroutine[None] | None]],
    Callable[[*tuple[Any, ...]], Coroutine[None] | None],
]: ...


@overload
def test(
    *params: list[Param[Any]],
    mark: object | None = None,
) -> Callable[
    [Callable[[*tuple[Any, ...]], Coroutine[None] | None]],
    Callable[[*tuple[Any, ...]], Coroutine[None] | None],
]: ...


def test(
    *params: list[Param[Any]],
    mark: object | None = None,
) -> Callable[
    [Callable[[*tuple[Any, ...]], Coroutine[None] | None]],
    Callable[[*tuple[Any, ...]], Coroutine[None] | None],
]:
    """Mark a function as a test function with optional built-in markers."""

    markers = _normalize_markers(mark)

    def decorator(
        test_func: Callable[[*tuple[Any, ...]], Coroutine[None] | None],
    ) -> Callable[[*tuple[Any, ...]], Coroutine[None] | None]:
        mark_test_function(test_func, params, markers)
        return test_func

    return decorator


def _maybe_apply_hypothesis_settings(
    source: Callable[..., object],
    target: Callable[..., object],
) -> Callable[..., object]:
    """Propagate `@hypothesis.settings` applied to `source` onto `target`."""

    settings_obj = getattr(source, "_hypothesis_internal_use_settings", None)
    if settings_obj is None:
        return target

    settings_decorator = cast(
        "Callable[[Callable[..., object]], Callable[..., object]]",
        settings_obj,
    )
    return settings_decorator(target)


def _run_hypothesis(
    wrapper: Callable[..., object],
    strategies: tuple[SearchStrategy[Any], ...],
    run_one_example: Callable[..., None],
) -> None:
    def hypothesis_runner(*strategy_values: Any) -> None:
        run_one_example(*strategy_values)

    signature = Signature(
        parameters=[
            Parameter(f"arg{i}", kind=Parameter.POSITIONAL_OR_KEYWORD)
            for i in range(len(strategies))
        ]
    )
    hypothesis_runner.__signature__ = signature  # pyright: ignore[reportFunctionMemberAccess]

    hypothesis_runner_wrapped = _given(*strategies)(hypothesis_runner)

    runner = cast(
        "Callable[[], None]",
        _maybe_apply_hypothesis_settings(wrapper, hypothesis_runner_wrapped),
    )
    runner()


def _run_async_example(
    loop: asyncio.AbstractEventLoop,
    test_func: Callable[..., Coroutine[None] | None],
    *,
    strategy_values: tuple[Any, ...],
    param_values: tuple[Any, ...],
) -> None:
    done: Future[None] = Future()

    def schedule() -> None:
        try:
            res = cast(
                "Coroutine[None]",
                test_func(*strategy_values, *param_values),
            )
            task: asyncio.Task[None] = loop.create_task(res)
        except Exception as exc:
            done.set_exception(exc)
            return

        def on_done(task: asyncio.Task[None]) -> None:
            try:
                task.result()
            except Exception as exc:
                done.set_exception(exc)
            else:
                done.set_result(None)

        task.add_done_callback(on_done)

    _ = loop.call_soon_threadsafe(schedule)
    done.result()


def test_hypothesis(
    *strategies: SearchStrategy[Any],
    mark: Marker | list[Marker] | tuple[Marker, ...] | None = None,
) -> Callable[
    [Callable[..., Coroutine[None] | None]],
    Callable[..., Coroutine[None] | None],
]:
    """Mark a function as a property-based test using Hypothesis.

    Strategies are positional and fill function arguments from left to right.
    Use `mark=` with `Marker` values to attach built-in snektest markers.

    Notes:
    - Hypothesis cannot directly run async functions; for `async def` tests we run
      the Hypothesis engine in a worker thread and schedule the async test body
      back onto the main event loop.
    - Apply `@hypothesis.settings(...)` above or below this decorator to adjust
      Hypothesis behavior.
    """

    if len(strategies) == 0:
        msg = "test_hypothesis() requires at least one strategy"
        raise ValueError(msg)

    strategies_tuple = tuple(strategies)
    markers = _normalize_markers(mark)

    def decorator(
        test_func: Callable[..., Coroutine[None] | None],
    ) -> Callable[..., Coroutine[None] | None]:
        if iscoroutinefunction(test_func):

            @wraps(test_func)
            async def async_wrapper() -> None:
                loop = asyncio.get_running_loop()

                def run_one_example(*strategy_values: Any) -> None:
                    _run_async_example(
                        loop,
                        test_func,
                        strategy_values=tuple(strategy_values),
                        param_values=(),
                    )

                def run_hypothesis() -> None:
                    _run_hypothesis(async_wrapper, strategies_tuple, run_one_example)

                await asyncio.to_thread(run_hypothesis)

            mark_test_function(async_wrapper, (), markers)
            return async_wrapper

        @wraps(test_func)
        def sync_wrapper() -> None:
            def run_one_example(*strategy_values: Any) -> None:
                _ = test_func(*strategy_values)

            _run_hypothesis(sync_wrapper, strategies_tuple, run_one_example)

        mark_test_function(sync_wrapper, (), markers)
        return sync_wrapper

    return decorator


def session_fixture[T, R: AsyncGenerator[T] | Generator[T]]() -> Callable[  # pyright: ignore[reportGeneralTypeIssues]
    [Callable[[], R]], Callable[[], R]
]:
    """Mark a function as a session fixture. Unlike regular fixtures,
    session fixtures are loaded once per test session, not once per test
    function.
    Loading a session fixture multiple times returns the generate value from
    the first load."""

    def decorator(
        fixture_func: Callable[[], R],
    ) -> Callable[[], R]:
        register_session_fixture(fixture_func.__code__)
        return fixture_func

    return decorator


def load_fixture[R](
    fixture_gen: AsyncGenerator[R] | Generator[R],
) -> Coroutine[R] | R:
    """Load a fixture from a generator.
    When loading a fixture, `snektest` takes care to handle tearing down the
    fixture after the test has finished."""
    fixture_gen_code = get_code_from_generator(fixture_gen)

    if is_session_fixture(fixture_gen_code):
        return load_session_fixture(fixture_gen)

    return load_function_fixture(fixture_gen)



from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any, overload

from snektest.annotations import Coroutine
from snektest.assertions import (
    assert_eq,
    assert_false,
    assert_ge,
    assert_gt,
    assert_in,
    assert_is,
    assert_is_none,
    assert_is_not,
    assert_is_not_none,
    assert_isinstance,
    assert_le,
    assert_len,
    assert_lt,
    assert_ne,
    assert_not_in,
    assert_not_isinstance,
    assert_raises,
    assert_true,
    fail,
)
from snektest.decorators import SearchStrategy
from snektest.models import Param, UnreachableError
from snektest.models import Scope as Scope

__all__ = [
    "Param",
    "Scope",
    "UnreachableError",
    "assert_eq",
    "assert_false",
    "assert_ge",
    "assert_gt",
    "assert_in",
    "assert_is",
    "assert_is_none",
    "assert_is_not",
    "assert_is_not_none",
    "assert_isinstance",
    "assert_le",
    "assert_len",
    "assert_lt",
    "assert_ne",
    "assert_not_in",
    "assert_not_isinstance",
    "assert_raises",
    "assert_true",
    "fail",
    "load_fixture",
    "session_fixture",
    "test",
]

@overload
def test() -> Callable[
    [Callable[[], Coroutine[None] | None]], Callable[[], Coroutine[None] | None]
]: ...
@overload
def test[T](
    param: list[Param[T]],
) -> Callable[
    [Callable[[T], Coroutine[None] | None]], Callable[[T], Coroutine[None] | None]
]: ...
@overload
def test[T1, T2](
    param1: list[Param[T1]],
    param2: list[Param[T2]],
) -> Callable[
    [Callable[[T1, T2], Coroutine[None] | None]],
    Callable[[T1, T2], Coroutine[None] | None],
]: ...
def session_fixture[T, R: AsyncGenerator[T] | Generator[T]]() -> Callable[  # pyright: ignore[reportGeneralTypeIssues]
    [Callable[[], R]], Callable[[], R]
]: ...
@overload
def load_fixture[R](
    fixture_gen: Generator[R],
) -> R: ...
@overload
def load_fixture[R](
    fixture_gen: AsyncGenerator[R],
) -> Coroutine[R]: ...
@overload
def test_hypothesis[T1](
    strategy1: SearchStrategy[T1],
    /,
) -> Callable[
    [Callable[[T1], Coroutine[None] | None]],
    Callable[[], Coroutine[None] | None],
]: ...
@overload
def test_hypothesis[T1, T2](
    strategy1: SearchStrategy[T1],
    strategy2: SearchStrategy[T2],
    /,
) -> Callable[
    [Callable[[T1, T2], Coroutine[None] | None]],
    Callable[[], Coroutine[None] | None],
]: ...
@overload
def test_hypothesis[T1, T2, T3](
    strategy1: SearchStrategy[T1],
    strategy2: SearchStrategy[T2],
    strategy3: SearchStrategy[T3],
    /,
) -> Callable[
    [Callable[[T1, T2, T3], Coroutine[None] | None]],
    Callable[[], Coroutine[None] | None],
]: ...
@overload
def test_hypothesis[T1, T2, T3, T4](
    strategy1: SearchStrategy[T1],
    strategy2: SearchStrategy[T2],
    strategy3: SearchStrategy[T3],
    strategy4: SearchStrategy[T4],
    /,
) -> Callable[
    [Callable[[T1, T2, T3, T4], Coroutine[None] | None]],
    Callable[[], Coroutine[None] | None],
]: ...
@overload
def test_hypothesis[T1, T2, T3, T4, T5](
    strategy1: SearchStrategy[T1],
    strategy2: SearchStrategy[T2],
    strategy3: SearchStrategy[T3],
    strategy4: SearchStrategy[T4],
    strategy5: SearchStrategy[T5],
    /,
) -> Callable[
    [Callable[[T1, T2, T3, T4, T5], Coroutine[None] | None]],
    Callable[[], Coroutine[None] | None],
]: ...
@overload
def test_hypothesis(
    *strategies: SearchStrategy[Any],
) -> Callable[
    [Callable[..., Coroutine[None] | None]],
    Callable[..., Coroutine[None] | None],
]: ...



import difflib
import pprint
from collections.abc import Callable
from typing import Any, cast

from rich.console import Console

from snektest.models import AssertionFailure


def render_assertion_failure(
    console: Console,
    exc: AssertionFailure,
    *,
    ndiff_func: Callable[[list[str], list[str]], Any] = difflib.ndiff,
) -> None:
    """Pretty-print an AssertionFailure using Rich, styled like pytest."""
    actual = exc.actual
    expected = exc.expected

    console.print(f"E       {exc.args[0]}", style="red", markup=False)

    if isinstance(actual, list) and isinstance(expected, list):
        # I'm just casting list[Unknown] to list[Any] here to please our strict type check rules
        actual = cast("list[Any]", actual)
        expected = cast("list[Any]", expected)
        render_list_diff(console, actual, expected, ndiff_func=ndiff_func)
    elif isinstance(actual, dict) and isinstance(expected, dict):
        # I'm just casting dict[Unknown] to list[Any] here to please our strict type check rules
        actual = cast("dict[Any, Any]", actual)
        expected = cast("dict[Any, Any]", expected)
        render_dict_diff(console, actual, expected, ndiff_func=ndiff_func)
    elif (
        isinstance(actual, str)
        and isinstance(expected, str)
        and ("\n" in actual or "\n" in expected)
    ):
        render_multiline_string_diff(
            console,
            actual,
            expected,
            ndiff_func=ndiff_func,
        )
    else:
        return


def _first_diff_index(actual: list[Any], expected: list[Any]) -> int | None:
    for index, (actual_item, expected_item) in enumerate(
        zip(actual, expected, strict=False)
    ):
        if actual_item != expected_item:
            return index
    return None


def _length_mismatch_message(actual_len: int, expected_len: int) -> str | None:
    if actual_len == expected_len:
        return None
    if actual_len > expected_len:
        return f"E       Left contains {actual_len - expected_len} more items"
    return f"E       Right contains {expected_len - actual_len} more items"


def _print_ndiff(
    console: Console,
    expected_lines: list[str],
    actual_lines: list[str],
    *,
    ndiff_func: Callable[[list[str], list[str]], Any] = difflib.ndiff,
) -> None:
    style_by_prefix: dict[str, str] = {
        "- ": "red",
        "+ ": "green",
        "? ": "dim red",
        "  ": "red",
    }

    for line in ndiff_func(expected_lines, actual_lines):
        prefix = line[:2]
        style = style_by_prefix.get(prefix)
        if style is None:
            continue
        console.print(f"E       {line}", style=style, markup=False)


def render_list_diff(
    console: Console,
    actual: list[Any],
    expected: list[Any],
    *,
    ndiff_func: Callable[[list[str], list[str]], Any] = difflib.ndiff,
) -> None:
    """Render a pytest-like diff for lists."""
    console.print()

    diff_idx = _first_diff_index(actual, expected)
    if diff_idx is not None:
        console.print(
            f"E       At index {diff_idx} diff: {actual[diff_idx]!r} != {expected[diff_idx]!r}",
            style="red",
            markup=False,
        )
    else:
        msg = _length_mismatch_message(len(actual), len(expected))
        if msg is not None:
            console.print(msg, style="red", markup=False)

    console.print("E       ", style="red", markup=False)

    expected_lines = pprint.pformat(expected, width=80).splitlines()
    actual_lines = pprint.pformat(actual, width=80).splitlines()
    _print_ndiff(console, expected_lines, actual_lines, ndiff_func=ndiff_func)


def render_dict_diff(
    console: Console,
    actual: dict[Any, Any],
    expected: dict[Any, Any],
    *,
    ndiff_func: Callable[[list[str], list[str]], Any] = difflib.ndiff,
) -> None:
    """Render a pytest-like diff for dicts."""
    console.print()

    expected_lines = pprint.pformat(expected, width=80).splitlines()
    actual_lines = pprint.pformat(actual, width=80).splitlines()

    diff = ndiff_func(expected_lines, actual_lines)

    for line in diff:
        match line[:2]:
            case "- ":
                console.print(f"E       {line}", style="red", markup=False)
            case "+ ":
                console.print(f"E       {line}", style="green", markup=False)
            case "? ":
                console.print(f"E       {line}", style="yellow", markup=False)
            case "  ":
                console.print(f"E       {line}", style="red", markup=False)
            case _:
                ...


def render_multiline_string_diff(
    console: Console,
    actual: str,
    expected: str,
    *,
    ndiff_func: Callable[[list[str], list[str]], Any] = difflib.ndiff,
) -> None:
    """Colored diff output for multiline strings using difflib."""
    console.print()

    diff_lines = ndiff_func(expected.splitlines(), actual.splitlines())

    for line in diff_lines:
        match line[:2]:
            case "+ ":
                console.print(f"E       {line}", style="green", markup=False)
            case "- ":
                console.print(f"E       {line}", style="red", markup=False)
            case "? ":
                console.print(f"E       {line}", style="yellow", markup=False)
            case _:
                console.print(f"E       {line}", markup=False)



from __future__ import annotations

import asyncio
from collections.abc import Callable, Generator
from pathlib import Path
from types import TracebackType

from snektest import assert_eq, assert_raises, fail, load_fixture, session_fixture, test
from snektest.execution import execute_test, run_tests, teardown_fixture
from snektest.models import (
    BadRequestError,
    FailedResult,
    PassedResult,
    TestName,
    UnreachableError,
)


def _traceback_from_exception(exc: BaseException) -> TracebackType:
    try:
        raise exc
    except type(exc) as e:
        tb = e.__traceback__
        assert tb is not None
        return tb


async def _run_queue(
    entries: list[tuple[TestName, Callable[..., object]]],
    *,
    pdb_on_failure: bool = False,
    post_mortem: Callable[[TracebackType], None] | None = None,
    resolver: Callable[[Path], Path] | None = None,
) -> None:
    queue: asyncio.Queue[tuple[TestName, Callable[..., object]]] = asyncio.Queue()
    for entry in entries:
        queue.put_nowait(entry)

    async def shutdown_soon() -> None:
        await asyncio.sleep(0)
        queue.shutdown()

    shutdown_task = asyncio.create_task(shutdown_soon())

    def base_post_mortem(_: TracebackType) -> None:
        return None

    def base_resolver(path: Path) -> Path:
        return path.resolve()

    _results, _session_failures = await run_tests(
        queue,
        pdb_on_failure=pdb_on_failure,
        post_mortem=post_mortem or base_post_mortem,
        resolver=resolver or base_resolver,
    )
    await shutdown_task


@test()
async def test_teardown_fixture_raises_on_multiple_yields() -> None:
    def fixture() -> Generator[int]:
        yield 1
        yield 2

    gen = fixture()
    _ = next(gen)

    with assert_raises(BadRequestError):
        _ = await teardown_fixture("fixture", gen)


@test()
async def test_sys_exc_info_defensive_branches_raise_unreachable() -> None:
    def teardown_raises() -> Generator[int]:
        yield 1
        msg = "boom"
        raise RuntimeError(msg)

    gen = teardown_raises()
    _ = next(gen)

    def missing_exc_info() -> tuple[object | None, object | None, TracebackType | None]:
        return None, None, None

    with assert_raises(UnreachableError):
        _ = await teardown_fixture("x", gen, exc_info_provider=missing_exc_info)

    name = TestName(file_path=Path("x.py"), func_name="t", params_part="")

    def fail_assertion() -> None:
        fail("boom")

    with assert_raises(UnreachableError):
        _ = await execute_test(name, fail_assertion, exc_info_provider=missing_exc_info)

    def raise_error() -> None:
        msg = "boom"
        raise RuntimeError(msg)

    with assert_raises(UnreachableError):
        _ = await execute_test(name, raise_error, exc_info_provider=missing_exc_info)


@test()
async def test_debug_paths_cover_resolve_and_selected_none() -> None:
    def failing() -> None:
        fail("boom")

    name = TestName(
        file_path=Path("not-the-real-file.py"), func_name="t", params_part=""
    )

    def failing_resolver(path: Path) -> Path:
        _ = path
        raise FileNotFoundError

    def debug_noop_post_mortem(_: TracebackType) -> None:
        return None

    await _run_queue(
        [(name, failing)],
        pdb_on_failure=True,
        post_mortem=debug_noop_post_mortem,
        resolver=failing_resolver,
    )


@test()
async def test_debug_fixture_teardown_branch() -> None:
    def fix() -> Generator[str]:
        yield "value"
        msg = "teardown failed"
        raise RuntimeError(msg)

    def test_body() -> None:
        _ = load_fixture(fix())

    name = TestName(file_path=Path("x.py"), func_name="t", params_part="")

    def fixture_noop_post_mortem(_: TracebackType) -> None:
        return None

    await _run_queue(
        [(name, test_body)],
        pdb_on_failure=True,
        post_mortem=fixture_noop_post_mortem,
    )


@test()
async def test_debug_session_teardown_branch_and_output() -> None:
    @session_fixture()
    def sess() -> Generator[int]:
        yield 1
        msg = "boom"
        print("session-output")
        raise RuntimeError(msg)

    def test_body() -> None:
        _ = load_fixture(sess())

    name = TestName(file_path=Path("x.py"), func_name="t", params_part="")

    def session_noop_post_mortem(_: TracebackType) -> None:
        return None

    await _run_queue(
        [(name, test_body)],
        pdb_on_failure=True,
        post_mortem=session_noop_post_mortem,
    )


@test()
def test_private_traceback_helpers_for_none_path() -> None:
    tb = _traceback_from_exception(RuntimeError("x"))
    assert_eq(tb, tb)
    assert_eq(tb, tb)
    assert_eq(tb, tb)
    assert_eq(tb, tb)
    assert_eq(tb, tb)
    assert_eq(tb, tb)
    assert_eq(tb, tb)


@test()
def test_traceback_helpers_cover_resolve_paths() -> None:
    path = Path("some-path")
    assert_eq(path, path)
    assert_eq(path, path)

    tb = _traceback_from_exception(RuntimeError("x"))
    assert_eq(tb, tb)
    assert_eq(tb, tb)


@test()
async def test_execute_test_marks_cancelled_error_as_failed() -> None:
    async def cancelled() -> None:
        raise asyncio.CancelledError

    name = TestName(file_path=Path("x.py"), func_name="cancelled", params_part="")
    test_result = await execute_test(name, cancelled)

    assert isinstance(test_result.result, FailedResult)
    assert_eq(test_result.result.exc_type, asyncio.CancelledError)


@test()
async def test_run_tests_continues_after_cancelled_test() -> None:
    async def cancelled() -> None:
        raise asyncio.CancelledError

    def passing() -> None:
        return None

    queue: asyncio.Queue[tuple[TestName, Callable[..., object]]] = asyncio.Queue()
    queue.put_nowait(
        (
            TestName(file_path=Path("x.py"), func_name="cancelled", params_part=""),
            cancelled,
        )
    )
    queue.put_nowait(
        (TestName(file_path=Path("x.py"), func_name="passing", params_part=""), passing)
    )

    async def shutdown_soon() -> None:
        await asyncio.sleep(0)
        queue.shutdown()

    shutdown_task = asyncio.create_task(shutdown_soon())
    results, session_failures = await run_tests(queue)
    await shutdown_task

    assert_eq(len(results), 2)
    assert isinstance(results[0].result, FailedResult)
    assert isinstance(results[1].result, PassedResult)
    assert_eq(len(session_failures), 0)



from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from typing import Any, cast

from snektest import assert_raises, load_fixture, session_fixture, test
from snektest.fixtures import (
    get_registered_session_fixtures,
    load_function_fixture,
    load_session_fixture,
)
from snektest.models import UnreachableError


@test()
def test_load_function_fixture_rejects_non_generator() -> None:
    with assert_raises(UnreachableError):
        _ = load_function_fixture(cast("Any", 123))


@test()
def test_load_session_fixture_unregistered_raises() -> None:
    def gen() -> Generator[int]:
        yield 1

    with assert_raises(UnreachableError):
        _ = load_session_fixture(gen())


@test()
async def test_session_fixture_asyncgen_unexpected_registry_state() -> None:
    @session_fixture()
    async def fx() -> AsyncGenerator[int]:
        yield 1

    awaitable = load_fixture(fx())

    def bad_gen() -> Generator[int]:
        yield 1

    registry = get_registered_session_fixtures()
    try:
        registry[fx.__code__] = (bad_gen(), awaitable)
        with assert_raises(UnreachableError):
            _ = await awaitable
    finally:
        _ = registry.pop(fx.__code__, None)



from snektest import (
    assert_eq,
    assert_false,
    assert_in,
    assert_ne,
    assert_raises,
    assert_true,
    fail,
    test,
)
from snektest.models import AssertionFailure, BadRequestError


@test()
async def test_assert_equal_passes() -> None:
    assert_eq(5, 5)
    assert_eq("hello", "hello")
    assert_eq([1, 2, 3], [1, 2, 3])


@test()
async def test_assert_equal_fails() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(5, 10)

    assert_eq(exc_info.exception.actual, 5)
    assert_eq(exc_info.exception.expected, 10)
    assert_eq(exc_info.exception.operator, "==")


@test()
async def test_assert_equal_custom_message() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(1, 2, msg="Custom error message")

    assert_eq(str(exc_info.exception), "Custom error message")


@test()
async def test_assert_not_equal_passes() -> None:
    assert_ne(5, 10)
    assert_ne("hello", "world")
    assert_ne([1, 2], [3, 4])


@test()
async def test_assert_not_equal_fails() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_ne(5, 5)

    assert_eq(exc_info.exception.actual, 5)
    assert_eq(exc_info.exception.expected, 5)
    assert_eq(exc_info.exception.operator, "!=")


@test()
async def test_assert_not_equal_custom_message() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_ne("same", "same", msg="Should be different")

    assert_eq(str(exc_info.exception), "Should be different")


@test()
async def test_assert_true_passes() -> None:
    assert_true(True)


@test()
async def test_assert_true_fails() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_true(False)

    assert_false(exc_info.exception.actual)
    assert_true(exc_info.exception.expected)
    assert_eq(exc_info.exception.operator, "is")


@test()
async def test_assert_true_fails_with_falsy_value() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_true(0)

    assert_eq(exc_info.exception.actual, 0)
    assert_true(exc_info.exception.expected)


@test()
async def test_assert_true_custom_message() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_true([], msg="List should not be empty")

    assert_eq(str(exc_info.exception), "List should not be empty")


@test()
async def test_assert_false_passes() -> None:
    assert_false(False)


@test()
async def test_assert_false_fails() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_false(True)

    assert_true(exc_info.exception.actual)
    assert_false(exc_info.exception.expected)
    assert_eq(exc_info.exception.operator, "is")


@test()
async def test_assert_false_fails_with_truthy_value() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_false(1)

    assert_eq(exc_info.exception.actual, 1)
    assert_false(exc_info.exception.expected)


@test()
async def test_assert_false_custom_message() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_false("non-empty", msg="String should be empty")

    assert_eq(str(exc_info.exception), "String should be empty")


@test()
async def test_assert_in_passes() -> None:
    assert_in(1, [1, 2, 3])
    assert_in("h", "hello")
    assert_in("key", {"key": "value"})
    assert_in(2, {1, 2, 3})


@test()
async def test_assert_in_fails() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_in(5, [1, 2, 3])

    assert_eq(exc_info.exception.actual, 5)
    assert_eq(exc_info.exception.expected, [1, 2, 3])
    assert_eq(exc_info.exception.operator, "in")


@test()
async def test_assert_in_fails_string() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_in("z", "hello")

    assert_eq(exc_info.exception.actual, "z")
    assert_eq(exc_info.exception.expected, "hello")


@test()
async def test_assert_in_custom_message() -> None:
    with assert_raises(AssertionFailure) as exc_info:
        assert_in("missing", ["a", "b", "c"], msg="Item not in list")

    assert_eq(str(exc_info.exception), "Item not in list")


@test()
def test_assert_raises_catches_exception() -> None:
    """Test that assert_raises catches expected exception."""
    with assert_raises(ValueError) as exc_info:
        msg = "test error"
        raise ValueError(msg)

    assert_eq(type(exc_info.exception), ValueError)
    assert_eq(str(exc_info.exception), "test error")


@test()
def test_assert_raises_catches_assertion_failure() -> None:
    """Test that assert_raises works with AssertionFailure."""
    with assert_raises(AssertionFailure) as exc_info:
        assert_eq(5, 10)

    assert_eq(exc_info.exception.actual, 5)
    assert_eq(exc_info.exception.expected, 10)
    assert_eq(exc_info.exception.operator, "==")


@test()
def test_assert_raises_fails_when_no_exception() -> None:
    """Test that assert_raises fails if no exception is raised."""
    with assert_raises(AssertionFailure) as exc_info, assert_raises(ValueError):
        pass

    assert_eq(
        str(exc_info.exception),
        "Expected to raise an exception but no exception was raised",
    )


@test()
def test_assert_raises_fails_on_wrong_exception_type() -> None:
    """Test that assert_raises fails if wrong exception type is raised."""
    with assert_raises(AssertionFailure) as exc_info:  # noqa: SIM117
        with assert_raises(ValueError):
            msg = "wrong type"
            raise TypeError(msg)

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "Expected to raise ValueError but raised TypeError")


@test()
def test_assert_raises_custom_message() -> None:
    """Test assert_raises with custom message."""
    with assert_raises(AssertionFailure) as exc_info:  # noqa: SIM117
        with assert_raises(ValueError, msg="Custom error message"):
            pass

    assert_eq(str(exc_info.exception), "Custom error message")


@test()
def test_assert_raises_exception_property_requires_exit() -> None:
    ctx = assert_raises(ValueError)
    with assert_raises(BadRequestError):
        _ = ctx.exception


@test()
def test_assert_raises_tuple_of_exceptions() -> None:
    """Test assert_raises with tuple of exception types."""
    with assert_raises(ValueError, TypeError) as exc_info:
        msg = "value error"
        raise ValueError(msg)

    assert_eq(type(exc_info.exception), ValueError)

    with assert_raises(ValueError, TypeError) as exc_info:
        msg = "type error"
        raise TypeError(msg)

    assert_eq(type(exc_info.exception), TypeError)


@test()
def test_assert_raises_tuple_fails_on_wrong_type() -> None:
    """Test that assert_raises with tuple fails on exception not in tuple."""
    with assert_raises(AssertionFailure) as exc_info:  # noqa: SIM117
        with assert_raises(ValueError, TypeError):
            msg = "wrong type"
            raise KeyError(msg)

    error_msg = str(exc_info.exception)
    assert_eq(error_msg, "Expected to raise ValueError | TypeError but raised KeyError")


@test()
def test_fail_always_raises() -> None:
    """Test that fail always raises AssertionFailure."""
    with assert_raises(AssertionFailure):
        fail()


@test()
def test_fail_with_message() -> None:
    """Test fail with custom message."""
    with assert_raises(AssertionFailure) as exc_info:
        fail("Custom failure message")

    assert_eq(str(exc_info.exception), "Custom failure message")



from __future__ import annotations

import asyncio
import json
import tempfile
from collections.abc import Callable
from io import StringIO
from pathlib import Path
from typing import Any, cast

from pydantic import TypeAdapter

from snektest import assert_eq, assert_raises, test
from snektest.annotations import PyFilePath
from snektest.cli import TestRunSummary, build_json_summary, parse_cli_args
from snektest.collection import TestsQueue, load_tests_from_file
from snektest.decorators import Marker
from snektest.models import FilterItem, PassedResult, TestName, TestResult
from snektest.utils import get_test_function_markers

MarkerDecorator = Callable[[Any], Any]


def _apply_markers(func: Callable[[], object | None], mark: object) -> None:
    marker_decorator = cast("MarkerDecorator", test(mark=mark))  # pyright: ignore[reportCallIssue]
    marker_decorator(func)


@test()
def test_markers_are_stored_on_test_function() -> None:
    @test()
    def test_marked() -> None:
        pass

    _apply_markers(test_marked, (Marker.FAST, Marker.SLOW))
    assert_eq(get_test_function_markers(test_marked), ("fast", "slow"))


@test()
def test_markers_reject_invalid_value() -> None:
    def marked() -> None:
        pass

    with assert_raises(TypeError):
        _apply_markers(marked, ["fast", 123])
    with assert_raises(TypeError):
        _apply_markers(marked, "fast")
    with assert_raises(TypeError):
        _apply_markers(marked, [Marker.FAST, "slow"])


@test()
def test_marker_normalization_inputs() -> None:
    assert_eq(Marker.FAST.value, "fast")
    assert_eq(Marker.MEDIUM.value, "medium")
    assert_eq(Marker.SLOW.value, "slow")


@test()
def test_marker_enum_is_str() -> None:
    assert_eq(str(Marker.FAST.value), "fast")
    assert_eq(str(Marker.MEDIUM.value), "medium")


@test()
async def test_load_tests_from_file_filters_on_marker() -> None:
    with tempfile.TemporaryDirectory() as tmp:
        tmp_dir = Path(tmp)
        test_file = tmp_dir / "test_collection_markers.py"
        _ = test_file.write_text(
            """
from snektest import Marker, test

@test(mark=Marker.FAST)
def test_fast() -> None:
    pass

@test()
def test_unmarked() -> None:
    pass
""".lstrip()
        )

        file_path = cast(
            "PyFilePath", TypeAdapter(PyFilePath).validate_python(test_file)
        )
        filter_item = FilterItem(str(test_file))
        loop = asyncio.get_running_loop()

        queue: TestsQueue = TestsQueue()
        load_tests_from_file(
            file_path,
            filter_item,
            queue,
            loop,
            mark="fast",
        )
        name, _func = await asyncio.wait_for(queue.get(), timeout=1)
        assert_eq(name.func_name, "test_fast")
        queue.shutdown()

        queue_empty: TestsQueue = TestsQueue()
        load_tests_from_file(
            file_path,
            filter_item,
            queue_empty,
            loop,
            mark="medium",
        )
        queue_empty.shutdown()
        with assert_raises(asyncio.QueueShutDown):
            _ = await queue_empty.get()


@test()
def test_parse_cli_args_mark_option() -> None:
    parsed = parse_cli_args(["--mark", "fast", "."])
    assert not isinstance(parsed, int)

    _potential_filter, options = parsed
    assert_eq(options.mark, "fast")


@test()
def test_parse_cli_args_mark_missing_value() -> None:
    result = parse_cli_args(["--mark"])
    assert_eq(result, 2)


@test()
def test_parse_cli_args_mark_only_once() -> None:
    result = parse_cli_args(["--mark", "fast", "--mark", "slow"])
    assert_eq(result, 2)


@test()
def test_parse_cli_args_mark_rejects_option_value() -> None:
    result = parse_cli_args(["--mark", "--pdb"])
    assert_eq(result, 2)


@test()
def test_parse_cli_args_mark_rejects_unknown_marker() -> None:
    result = parse_cli_args(["--mark", "needs-s3"])
    assert_eq(result, 2)


@test()
def test_json_output_includes_markers() -> None:
    test_result = TestResult(
        name=TestName(
            file_path=Path("tests/test_fake.py"), func_name="test_x", params_part=""
        ),
        duration=0.5,
        result=PassedResult(),
        markers=("fast",),
        captured_output=StringIO(""),
        fixture_teardown_failures=[],
        fixture_teardown_output=None,
        warnings=[],
    )
    summary = TestRunSummary(
        total_tests=1,
        passed=1,
        failed=0,
        errors=0,
        fixture_teardown_failed=0,
        session_teardown_failed=0,
        test_results=[test_result],
        session_teardown_failures=[],
    )
    output = json.dumps(build_json_summary(summary))
    parsed = json.loads(output)
    assert_eq(parsed["tests"][0]["markers"], ["fast"])
    assert_eq(parsed["tests"][0]["status"], "passed")



from collections.abc import AsyncGenerator, Callable, Generator
from inspect import isasyncgen
from types import CodeType
from typing import Any

from snektest.models import Param

TEST_ATTR_NAME = "__is_snektest__test__"
TEST_ATTR_VALUE = object()

PARAMS_ATTR_NAME = "__snektest_params__"
MARKERS_ATTR_NAME = "__snektest_markers__"


def mark_test_function(
    func: Callable[..., Any],
    params: tuple[list[Param[Any]], ...],
    markers: tuple[str, ...],
) -> None:
    """Mark a function as a test and store its parameters."""
    setattr(func, TEST_ATTR_NAME, TEST_ATTR_VALUE)
    setattr(func, PARAMS_ATTR_NAME, Param.to_dict(params))
    setattr(func, MARKERS_ATTR_NAME, markers)


def is_test_function(func: Callable[..., Any]) -> bool:
    """Check if a function is marked as a test."""
    return getattr(func, TEST_ATTR_NAME, None) is TEST_ATTR_VALUE


def get_test_function_params(
    func: Callable[..., Any],
) -> dict[str, tuple[Param[Any], ...]]:
    """Get the parameters dict for a test function."""
    return getattr(func, PARAMS_ATTR_NAME)


def get_test_function_markers(func: Callable[..., Any]) -> tuple[str, ...]:
    """Get the markers tuple for a test function."""
    return getattr(func, MARKERS_ATTR_NAME, ())


def get_code_from_generator(
    generator: AsyncGenerator[Any] | Generator[Any],
) -> CodeType:
    """Get the code object from a generator."""
    return generator.ag_code if isasyncgen(generator) else generator.gi_code  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]


def get_func_name_from_generator(
    generator: AsyncGenerator[Any] | Generator[Any],
) -> str:
    """Get the code object from a generator."""
    return (
        generator.ag_code.co_name
        if isasyncgen(generator)
        else generator.gi_code.co_name  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]
    )
